from PyQt5 import QtWidgets

from PyQt5.QtCore import Qt
from .gui_fit_widget import Ui_Form

from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QMessageBox, QLineEdit, QCheckBox, QFileDialog
import numpy as np

from spectramanipulator.settings import Settings
from spectramanipulator.logger import Logger
from spectramanipulator.treeview.item import SpectrumItem, SpectrumItemGroup
from spectramanipulator.console import Console
from spectramanipulator.spectrum import fi
import spectramanipulator

import pyqtgraph as pg

# from user_namespace import UserNamespace

from lmfit import Minimizer, Parameters
from scipy.integrate import odeint

import spectramanipulator.fitting.fitmodels as fitmodels
import spectramanipulator.fitting.userfitmodels as userfitmodels
import inspect
from ..general_model import GeneralModel

from ..plotwidget import PlotWidget
from .fitresult import FitResult

from ..utils.syntax_highlighter import PythonHighlighter, KineticModelHighlighter

import glob
import os

from ..misc import int_default_color_scheme


class FitWidget(QtWidgets.QWidget, Ui_Form):
    # static variables
    is_opened = False
    _instance = None

    # maximum number of parameters
    max_count = 80

    def __init__(self, dock_widget, accepted_func, spectrum=None, parent=None):
        super(FitWidget, self).__init__(parent)

        # if FitWidget._instance is not None:
        #     PlotWidget._instance.removeItem(FitWidget._instance.lr)

        self.setupUi(self)

        self.dock_widget = dock_widget
        self.accepted_func = accepted_func

        self.fitted_params = None
        self.covariance_matrix = None
        self.errors = None
        self.fitted_spectrum = None
        self.residual_spectrum = None

        self.plot_fit = None
        self.plot_residuals = None

        # if spectrum is None - we have just open the widget for function plotting
        self.spectrum = spectrum

        # self.plot_widget = PlotWidget._instance

        self.cbUpdateInitValues.setCheckState(Qt.Checked)

        # x0, x1 = PlotWidget.get_view_range()

        self.lr = PlotWidget.add_linear_region(bounds=(self.spectrum.x.min(), self.spectrum.x.max())) if \
                self.spectrum is not None else None

        # self.lr = pg.LinearRegionItem([x0 + (1 - f) * xy_dif, x0 + f * xy_dif],
        #                               brush=QtGui.QBrush(QtGui.QColor(0, 255, 0, 20)),
        #                               bounds=(self.spectrum.x.min(), self.spectrum.x.max())
        #                               if self.spectrum is not None else None)
        # update region when the focus is lost and when the user presses enter
        self.leX0.returnPressed.connect(self.updateRegion)
        self.leX1.returnPressed.connect(self.updateRegion)
        self.leX0.focus_lost.connect(self.updateRegion)
        self.leX1.focus_lost.connect(self.updateRegion)

        self.lr.sigRegionChanged.connect(self.updatePlot)
        self.lr.sigRegionChangeFinished.connect(self.update_initial_values)
        self.updatePlot()

        self.sbParamsCount.valueChanged.connect(self.params_count_changed)
        self.btnFit.clicked.connect(self.fit)
        self.btnSimulateModel.clicked.connect(self.plot_function)
        self.btnPrintReport.clicked.connect(self.print_report)
        self.btnClearPlot.clicked.connect(self.clear_plot)
        # self.btnPrintDiffEquations.clicked.connect(self.print_diff_eq)
        self.btnOK.clicked.connect(self.accept)
        self.btnCancel.clicked.connect(self.reject)
        self.cbModel.currentIndexChanged.connect(self.model_changed)
        self.cbGenModels.currentIndexChanged.connect(self.cbGenModel_changed)
        self.btnBuildModel.clicked.connect(lambda: self.build_gen_model(True))
        self.cbShowBackwardRates.stateChanged.connect(self.cbShowBackwardRates_checked_changed)
        self.btnSaveCustomModel.clicked.connect(self.save_general_model_as)

        self.sbParamsCount.setMaximum(self.max_count)

        self.pteEquation_highlighter = PythonHighlighter(self.pteEquation.document())
        self.ppteScheme_highlighter = KineticModelHighlighter(self.pteScheme.document())

        self.params_list = []
        self.lower_bound_list = []
        self.value_list = []
        self.upper_bound_list = []
        self.fixed_list = []
        self.error_list = []

        for i in range(self.max_count):
            self.params_list.append([QLineEdit(), QLineEdit()])
            self.lower_bound_list.append([QLineEdit(), QLineEdit()])
            self.value_list.append([QLineEdit(), QLineEdit()])
            self.upper_bound_list.append([QLineEdit(), QLineEdit()])
            self.fixed_list.append([QCheckBox(), QCheckBox()])
            self.error_list.append([QLineEdit(), QLineEdit()])

            self.params_list[i][1].setReadOnly(True)
            self.error_list[i][0].setReadOnly(True)
            self.error_list[i][1].setReadOnly(True)

            for j, layout in enumerate([self.predefGridLayout, self.cusModelGridLayout]):
                layout.addWidget(self.params_list[i][j], i + 1, 0, 1, 1)
                layout.addWidget(self.lower_bound_list[i][j], i + 1, 1, 1, 1)
                layout.addWidget(self.value_list[i][j], i + 1, 2, 1, 1)
                layout.addWidget(self.upper_bound_list[i][j], i + 1, 3, 1, 1)
                layout.addWidget(self.fixed_list[i][j], i + 1, 4, 1, 1)
                layout.addWidget(self.error_list[i][j], i + 1, 5, 1, 1)

            self.fixed_list[i][0].stateChanged.connect(self.fixed_checked_changed)
            self.fixed_list[i][1].stateChanged.connect(self.fixed_checked_changed)

            # if we are only plotting functions, disable some controls
            if self.spectrum is None:
                for j in (0, 1):
                    self.lower_bound_list[i][j].setEnabled(False)
                    self.upper_bound_list[i][j].setEnabled(False)
                    self.fixed_list[i][j].setEnabled(False)
                    self.error_list[i][j].setEnabled(False)

        if self.spectrum is None:
            self.btnFit.setEnabled(False)
            self.btnPrintReport.setEnabled(False)
            self.cbMethod.setEnabled(False)
            self.cbUpdateInitValues.setEnabled(False)

        # get all models from fitmodels, get classes that inherits from Model base class and sort them by name
        # and number of species
        classes = inspect.getmembers(sys.modules[fitmodels.__name__], inspect.isclass)
        tuples = filter(lambda tup: issubclass(tup[1], fitmodels._Model) and tup[1] is not fitmodels._Model, classes)
        # same load user defined fit models
        classes_usr = inspect.getmembers(sys.modules[userfitmodels.__name__], inspect.isclass)
        tuples_usr = filter(lambda tup: issubclass(tup[1], fitmodels._Model) and tup[1] is not fitmodels._Model,
                            classes_usr)
        # load models
        self.models = sorted(list(map(lambda tup: tup[1], tuples)) + list(map(lambda tup: tup[1], tuples_usr)),
                             key=lambda cls: cls.name)
        # fill the available models combo box with model names
        self.cbModel.addItems(map(lambda m: m.name, self.models))

        self.gen_models_paths = []
        self.update_general_models()

        # abbr is name that is needed for lmfit.fitting method
        self.methods = [
            {'name': 'Trust Region Reflective method', 'abbr': 'least_squares'},
            {'name': 'Levenberg–Marquardt', 'abbr': 'leastsq'},
            {'name': 'Differential evolution', 'abbr': 'differential_evolution'},
            {'name': 'Nelder-Mead, Simplex method (no error)', 'abbr': 'nelder'},
            {'name': 'L-BFGS-B (no error)', 'abbr': 'lbfgsb'},
            {'name': 'Powell (no error)', 'abbr': 'powell'}
        ]

        self.cbMethod.addItems(map(lambda m: m['name'], self.methods))

        self.cbCustom.stateChanged.connect(self.cbCustom_checked_changed)
        self.cbCustom_checked_changed()
        self.model_changed()

        self.current_model = None
        self.current_general_model = None
        self.general_model_params = None
        self.fit_result = None
        self.plotted_functions = []
        self.plotted_function_spectra = []

        self.accepted = False
        FitWidget.is_opened = True
        FitWidget._instance = self

        self.dock_widget.parent().resizeDocks([self.dock_widget], [250], Qt.Vertical)
        self.dock_widget.titleBarWidget().setText(
            "Fit of {}".format(self.spectrum.name) if self.spectrum is not None else "Function plotter")
        self.dock_widget.setWidget(self)
        self.dock_widget.setVisible(True)

        self.set_lower_upper_enabled()

    # @staticmethod
    # def get_instance():
    #     return FitWidget._instance

    def update_general_models(self):
        # curr_idx = self.cbGenModels.currentIndex() if len(self.cbGenModels) > 0 else None
        self.cbGenModels.clear()
        self.gen_models_paths = []

        module_path = os.path.abspath(spectramanipulator.__file__)  # get the absolute path of a module
        fpath = os.path.join(os.path.dirname(module_path), Settings.general_models_dir, '*.json')

        for fpath in glob.glob(fpath, recursive=True):
            fname = os.path.splitext(os.path.split(fpath)[1])[0]
            self.gen_models_paths.append(fpath)
            self.cbGenModels.addItem(fname)
        #
        # if curr_idx is not None:
        #     self.cbGenModels.setCurrentIndex(curr_idx)

    @staticmethod
    def check_state(checked):
        return Qt.Checked if checked else 0

    def updateRegion(self):
        try:
            x0, x1 = float(self.leX0.text()), float(self.leX1.text())
            self.lr.setRegion((x0, x1))
        except ValueError:
            pass

    def save_general_model_as(self):
        if self.current_general_model is None:
            return
        curr_model_path = self.gen_models_paths[self.cbGenModels.currentIndex()]
        fil = "Json (*.json)"

        filepath = QFileDialog.getSaveFileName(caption="Save Custom Kinetic Model",
                                               directory=curr_model_path,
                                               filter=fil, initialFilter=fil)
        if filepath[0] == '':
            return

        try:
            init, _, rates, _ = self.get_params_from_fields()
            self.current_general_model.scheme = self.pteScheme.toPlainText()
            self.current_general_model.set_rates(rates)
            self.current_general_model.initial_conditions = dict(
                zip(self.current_general_model.get_compartments(), init))
            self.current_general_model.save(filepath[0])
        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.critical(self, 'Saving error', 'Error, first build the model.', QMessageBox.Ok)
            return

        self.update_general_models()
        k = 0
        for k, fpath in enumerate(self.gen_models_paths):
            fname = os.path.splitext(os.path.split(fpath)[1])[0]
            if fname == os.path.splitext(os.path.split(filepath[0])[1])[0]:
                break

        # set saved model as current index
        self.cbGenModels.setCurrentIndex(k)

    def get_params_from_fields(self):
        if self.current_general_model is None:
            return

        n_comps = len(self.current_general_model.get_compartments())
        n_params = len(self.current_general_model.elem_reactions)

        init_cond = np.empty(n_comps, dtype=np.float64)
        coefs = np.empty(n_comps, dtype=np.float64)
        rates = np.zeros((n_params, 2), dtype=np.float64)

        for i, (i_idx, coef_idx) in enumerate(zip(range(n_comps), range(n_comps, 2*n_comps))):
            init_cond[i] = float(self.value_list[i_idx][1].text())
            coefs[i] = float(self.value_list[coef_idx][1].text())

        show_bw_rates = self.cbShowBackwardRates.isChecked()
        last_idx = 2*n_comps + n_params * (2 if show_bw_rates else 1)
        forward_idxs = range(2*n_comps, last_idx, (2 if show_bw_rates else 1))

        for i, idx in enumerate(forward_idxs):
            rates[i, 0] = float(self.value_list[idx][1].text())
            if show_bw_rates:
                rates[i, 1] = float(self.value_list[idx + 1][1].text())

        return init_cond, coefs, rates, float(self.value_list[last_idx][1].text())

    def get_values_from_params(self, params):

        n_comps = len(self.current_general_model.get_compartments())
        n_params = len(self.current_general_model.elem_reactions)

        init_cond = np.empty(n_comps, dtype=np.float64)
        coefs = np.empty(n_comps, dtype=np.float64)
        rates = np.zeros((n_params, 2), dtype=np.float64)

        l_params = list(params.values())

        for i, (i_idx, coef_idx) in enumerate(zip(range(n_comps), range(n_comps, 2*n_comps))):
            init_cond[i] = l_params[i_idx].value
            coefs[i] = l_params[coef_idx].value

        show_bw_rates = self.cbShowBackwardRates.isChecked()
        forward_idxs = range(2*n_comps, 2*n_comps + n_params * (2 if show_bw_rates else 1), (2 if show_bw_rates else 1))

        for i, idx in enumerate(forward_idxs):
            rates[i, 0] = l_params[idx].value
            if show_bw_rates:
                rates[i, 1] = l_params[idx + 1].value

        return init_cond, coefs, rates, l_params[-1]  # last is intercept

    def updatePlot(self):
        x0, x1 = self.lr.getRegion()
        self.leX0.setText("{:.4g}".format(x0))
        self.leX1.setText("{:.4g}".format(x1))

    def cbGenModel_changed(self):
        self.current_general_model = GeneralModel.load(self.gen_models_paths[self.cbGenModels.currentIndex()])
        self.pteScheme.setPlainText(self.current_general_model.scheme)
        self.build_gen_model(load_scheme=False)

    def cbShowBackwardRates_checked_changed(self):
        if self.current_general_model is None:
            return

        self.build_gen_model(load_scheme=False)

    def build_gen_model(self, load_scheme=True):
        if load_scheme:
            try:
                self.current_general_model = GeneralModel.from_text(self.pteScheme.toPlainText())
            except Exception as e:
                Logger.message(e.__str__())
                QMessageBox.critical(self, 'Build failed', f'Invalid model:\n\n{e.__str__()}', QMessageBox.Ok)
                return

        self.current_general_model.build_func()

        comps = self.current_general_model.get_compartments()
        init_comps_names = list(map(lambda n: f'[{n}]0', comps))
        rates = self.current_general_model.get_rates(self.cbShowBackwardRates.isChecked(),
                                                     append_values=True)

        n_params = 2 * len(comps) + len(rates) + 1  # + intercept
        self.set_field_visible(n_params, 1)

        self.general_model_params = Parameters()

        k = 0
        # initial conditions
        for init in self.current_general_model.initial_conditions.values():
            self.general_model_params.add(f'param{k}', value=init,
                                          vary=False,
                                          min=0,
                                          max=np.inf)
            k += 1
        # coefficients
        for i in range(len(comps)):
            self.general_model_params.add(f'param{k}', value=1 if i == 0 else 0,
                                          vary=True if i == 0 else False,
                                          min=-np.inf,
                                          max=np.inf)
            k += 1

        # rates
        rates_names = []
        for name, value in rates:
            rates_names.append(name)
            self.general_model_params.add(f'param{k}', value=value,
                                          vary=True,
                                          min=0,
                                          max=np.inf)
            k += 1

        # add intercept
        self.general_model_params.add('y0', value=0, vary=True, min=-np.inf, max=np.inf)

        for i, (p, name) in enumerate(zip(self.general_model_params.values(),
                                          init_comps_names + comps + rates_names + ['y0'])):
            self.params_list[i][1].setText(name)
            self.value_list[i][1].setText(f"{p.value:.5g}")
            self.lower_bound_list[i][1].setText(str(p.min))
            self.upper_bound_list[i][1].setText(str(p.max))
            self.fixed_list[i][1].setChecked(not p.vary)

        self.set_lower_upper_enabled()

    def cbCustom_checked_changed(self):
        if self.cbCustom.isChecked():
            self.sbParamsCount.setEnabled(True)
            self.pteEquation.setReadOnly(False)
            for i in range(self.max_count):
                self.params_list[i][0].setReadOnly(False)
        else:
            self.sbParamsCount.setEnabled(False)
            self.pteEquation.setReadOnly(True)
            for i in range(self.max_count):
                self.params_list[i][0].setReadOnly(True)

    def update_initial_values(self):

        if not self.cbUpdateInitValues.isChecked():
            return

        if self.current_model is None:
            return

        if self.spectrum is None:
            # iterate parameters and fill the values
            for i, p in enumerate(self.current_model.params.values()):
                self.value_list[i][0].setText("{:.4g}".format(p.value))
            return

        x0, x1 = self.lr.getRegion()

        start_idx = fi(self.spectrum.data[:, 0], x0)
        end_idx = fi(self.spectrum.data[:, 0], x1) + 1

        x_data = self.spectrum.data[start_idx:end_idx, 0]
        y_data = self.spectrum.data[start_idx:end_idx, 1]

        self.current_model.initialize_values(x_data, y_data)

        # iterate parameters and fill the values
        for i, p in enumerate(self.current_model.params.values()):
            self.value_list[i][0].setText("{:.4g}".format(p.value))

    def model_changed(self):
        # initialize new model
        self.current_model = self.models[self.cbModel.currentIndex()]()
        params_count = self.current_model.par_count()
        self.sbParamsCount.setValue(params_count)
        self.pteEquation.setPlainText(self.current_model.get_func_string())

        # iterate parameters and fill the fields
        for i, p in enumerate(self.current_model.params.values()):
            self.params_list[i][0].setText(p.name)
            self.lower_bound_list[i][0].setText(str(p.min))
            self.upper_bound_list[i][0].setText(str(p.max))
            self.fixed_list[i][0].setChecked(not p.vary)

        self.update_initial_values()

    def clear_plot(self, keep_spectra=False):
        if self.plot_fit is not None and self.plot_residuals is not None:
            self.plot_widget.plotItem.removeItem(self.plot_fit)
            self.plot_widget.plotItem.removeItem(self.plot_residuals)
            try:
                self.plot_widget.legend.remove_last_item()
                self.plot_widget.legend.remove_last_item()
            except IndexError:
                pass

            self.plot_fit = None
            self.plot_residuals = None

        if len(self.plotted_functions) > 0:
            for item in self.plotted_functions:
                self.plot_widget.plotItem.removeItem(item)
            try:
                for i in range(len(self.plotted_functions)):
                    self.plot_widget.legend.remove_last_item()
            except IndexError:
                pass

            self.plotted_functions = []
            if not keep_spectra:
                self.plotted_function_spectra = []

    def fit(self):

        try:
            self._fit()
        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.warning(self, 'Fitting Error', e.__str__(), QMessageBox.Ok)

    # def remove_last_fit(self):
    #     if self.plot_fit is not None and self.plot_residuals is not None:
    #         self.plot_widget.plotItem.removeItem(self.plot_fit)
    #         self.plot_widget.plotItem.removeItem(self.plot_residuals)
    #         try:
    #             self.plot_widget.legend.removeLastItem()
    #             self.plot_widget.legend.removeLastItem()
    #         except IndexError:
    #             pass

    def _setup_model(self):
        # custom model check box checked, we have to create own model from scratch
        if self.cbCustom.isChecked():

            self.current_model = fitmodels._Model()

            p_names = []

            for i in range(int(self.sbParamsCount.value())):
                param = self.params_list[i].text()
                p_names.append(param)
                self.current_model.params.add(param)

            # create new function, fill params names after x and body of the function (take care of proper indentation)
            func_command = "def user_defined_func(x, {}):\n\t{}".format(','.join(p_names),
                                                                        self.pteEquation.toPlainText().replace('\n',
                                                                                                               '\n\t'))
            exec(func_command, locals(), globals())  # locals and globals had to be changed

            # change the blank function in abstract model by user defined
            self.current_model.func = user_defined_func

        if self.current_model is None:
            self.current_model = self.models[self.cbModel.currentIndex()]()

    def _fit(self):
        import numpy as np

        x0, x1 = self.lr.getRegion()

        start_idx = fi(self.spectrum.data[:, 0], x0)
        end_idx = fi(self.spectrum.data[:, 0], x1) + 1

        x_data = self.spectrum.data[start_idx:end_idx, 0]
        y_data = self.spectrum.data[start_idx:end_idx, 1]

        tab_idx = self.tabWidget.currentIndex()

        if tab_idx == 0:
            self._setup_model()

        # fill the parameters from fields
        for i, p in enumerate((self.current_model.params if tab_idx == 0 else self.general_model_params).values()):
            p.value = float(self.value_list[i][tab_idx].text())
            p.min = float(self.lower_bound_list[i][tab_idx].text())
            p.max = float(self.upper_bound_list[i][tab_idx].text())
            p.vary = not self.fixed_list[i][tab_idx].isChecked()

        def y_fit(params):
            if tab_idx == 0:
                y = self.current_model.wrapper_func(x_data, params)
            else:
                init, coefs, rates, y0 = self.get_values_from_params(params)
                sol = self._simul_custom_model(init, rates, x_data)
                y = (coefs * sol).sum(axis=1, keepdims=False) + y0
            return y

        def residuals(params):
            y = y_fit(params)
            e = y - y_data
            if self.cbPropWeighting.isChecked():
                e /= y * y

            return e

        minimizer = Minimizer(residuals, self.current_model.params if tab_idx == 0 else self.general_model_params)

        method = self.methods[self.cbMethod.currentIndex()]['abbr']
        result = minimizer.minimize(method=method)  # fit

        if tab_idx == 0:
            self.current_model.params = result.params
        else:
            self.general_model_params = result.params

        # fill fields
        values_errors = np.zeros((len(result.params), 2), dtype=np.float64)
        for i, p in enumerate(result.params.values()):
            values_errors[i, 0] = p.value
            values_errors[i, 1] = p.stderr if p.stderr is not None else 0

            self.value_list[i][tab_idx].setText(f"{p.value:.4g}")
            self.error_list[i][tab_idx].setText(f"{p.stderr:.4g}" if p.stderr else '')

        y_fit_data = y_fit(result.params)
        y_residuals = y_data - y_fit_data

        # self.remove_last_fit()
        self.clear_plot()

        self.plot_fit = self.plot_widget.plotItem.plot(x_data, y_fit_data,
                                                       pen=pg.mkPen(color=QColor(0, 0, 0, 200), width=2.5),
                                                       name="Fit of {}".format(self.spectrum.name))

        self.plot_residuals = self.plot_widget.plotItem.plot(x_data, y_residuals,
                                                             pen=pg.mkPen(color=QColor(255, 0, 0, 150), width=1),
                                                             name="Residuals of {}".format(self.spectrum.name))

        self.fitted_spectrum = SpectrumItem.from_xy_values(x_data, y_fit_data,
                                                       name="Fit of {}".format(self.spectrum.name),
                                                       color='black', line_width=2.5, line_type=Qt.SolidLine)

        self.residual_spectrum = SpectrumItem.from_xy_values(x_data, y_residuals,
                                                         name="Residuals of {}".format(self.spectrum.name),
                                                         color='red', line_width=1, line_type=Qt.SolidLine)

        self.fit_result = FitResult(result, minimizer, values_errors, (self.current_model if tab_idx == 0 else self.current_general_model),
                                    data_item=self.spectrum, fit_item=self.fitted_spectrum,
                                    residuals_item=self.residual_spectrum)

    def plot_function(self):

        # self._plot_function()
        try:

            self._plot_function()
        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.warning(self, 'Plotting Error', e.__str__(), QMessageBox.Ok)

    def _simul_custom_model(self, j, rates, x_data):
        if x_data[0] > 0:  # initial conditions are valid for time=0
            n = 100  # prepend x values with 100 points if not starting with zero time
            x_prepended = np.concatenate((np.linspace(0, x_data[0], n, endpoint=False), x_data))
            return odeint(self.current_general_model.func, j, x_prepended, args=(rates,))[n:, :]

        elif x_data[0] < 0:
            x_pos = x_data[x_data >= 0]  # find x >= 0
            sol = np.zeros((x_data.shape[0], j.shape[0]), dtype=np.float64)
            if x_pos.shape[0] > 1:  # simulate only for at least 2 positive values
                sol[(x_data < 0).sum():, :] = self._simul_custom_model(j, rates, x_pos)  # use recursion here

            return sol

        # for x_data[0] == 0
        return odeint(self.current_general_model.func, j, x_data, args=(rates,))

    def _plot_function(self):
        x0, x1 = self.lr.getRegion()

        # if this is a fit widget with real spectrum, use x range from that spectrum,
        # otherwise, use np.linspace with defined number of points in Settings
        if self.spectrum is not None:
            start_idx = fi(self.spectrum.data[:, 0], x0)
            end_idx = fi(self.spectrum.data[:, 0], x1) + 1
            x_data = self.spectrum.data[start_idx:end_idx, 0]
        else:
            x_data = np.linspace(x0, x1, num=Settings.FP_num_of_points)

        tab_idx = self.tabWidget.currentIndex()

        if tab_idx == 0:  # equation-based model
            self._setup_model()

            par_count = self.current_model.par_count()

            # get params from field to current model
            for i in range(par_count):
                param = self.params_list[i][0].text()
                self.current_model.params[param].value = float(self.value_list[i][0].text())

            # calculate the y values according to our model
            y_data = self.current_model.wrapper_func(x_data, self.current_model.params)
        else:  # custom kinetic model
            if self.current_general_model is None:
                self.cbGenModel_changed()

            init, coefs, rates, y0 = self.get_params_from_fields()
            par_count = 2 * init.shape[0] + rates.shape[0]
            n_params = len(init)

            sol = self._simul_custom_model(init, rates, x_data)

            if self.cbPlotAllComps.isChecked():
                spectra = SpectrumItemGroup(name='all comparments plot')
                for i in range(n_params):
                    name = self.params_list[i+n_params][tab_idx].text()
                    self.plotted_functions.append(self.plot_widget.plotItem.plot(x_data, sol[:, i],
                                                                             pen=pg.mkPen(color=int_default_color_scheme(i),
                                                                                          width=1),
                                                                             name=name))
                    spectra.children.append(SpectrumItem.from_xy_values(x_data, sol[:, i], name=name))
                self.plotted_function_spectra.append(spectra)
                return

            y_data = (sol * coefs).sum(axis=1, keepdims=False) + y0

        params = ', '.join([f"{self.params_list[i][tab_idx].text()}={self.value_list[i][tab_idx].text()}" for i in range(par_count)])

        name = f"Func plot: {params}"
        self.plotted_functions.append(self.plot_widget.plotItem.plot(x_data, y_data,
                                                                     pen=pg.mkPen(color=QColor(0, 0, 0, 255),
                                                                                  width=1),
                                                                     name=name))

        self.plotted_function_spectra.append(SpectrumItem.from_xy_values(x_data, y_data, name=name))

    def set_lower_upper_enabled(self):
        if self.spectrum is None:
            return
        for i in range(len(self.fixed_list)):
            enabled_0 = not self.fixed_list[i][0].isChecked()
            enabled_1 = not self.fixed_list[i][1].isChecked()

            self.lower_bound_list[i][0].setEnabled(enabled_0)
            self.upper_bound_list[i][0].setEnabled(enabled_0)

            self.lower_bound_list[i][1].setEnabled(enabled_1)
            self.upper_bound_list[i][1].setEnabled(enabled_1)

    def fixed_checked_changed(self, value):
        if self.spectrum is None:
            return

        checkbox = self.sender()  # get the checkbox that was clicked on

        idx = self.tabWidget.currentIndex()

        i = 0
        for i, (ch1, ch2) in enumerate(self.fixed_list):
            if ch1 == checkbox or ch2 == checkbox:
                break

        enabled = True if value == Qt.Unchecked else False

        self.lower_bound_list[i][idx].setEnabled(enabled)
        self.upper_bound_list[i][idx].setEnabled(enabled)

    def params_count_changed(self):

        count = int(self.sbParamsCount.value())

        self.set_field_visible(count, 0)

    def set_field_visible(self, n_params, idx=0):
        for i in range(self.max_count):
            visible = n_params > i

            self.params_list[i][idx].setVisible(visible)
            self.lower_bound_list[i][idx].setVisible(visible)
            self.value_list[i][idx].setVisible(visible)
            self.upper_bound_list[i][idx].setVisible(visible)
            self.fixed_list[i][idx].setVisible(visible)
            self.error_list[i][idx].setVisible(visible)

    def set_result(self):
        pass

    def print_report(self):
        if self.fit_result is None:
            return
        try:
            report = "Fitting of '{}' succeeded. " \
                     "You can find fit results in variable " \
                     "'fit'.\n{}\n".format(self.spectrum.name,
                                           self.fit_result.report(print=False))

            Logger.console_message(report)
            Console.push_variables({'fit': self.fit_result})

        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.warning(self, 'Printing Error', e.__str__(), QMessageBox.Ok)

    def accept(self):
        # self.remove_last_fit()
        self.clear_plot(keep_spectra=True)
        self.print_report()

        self.accepted = True
        FitWidget.is_opened = False
        FitWidget._instance = None
        # self.plot_widget.removeItem(self.lr)
        # del self.lr
        PlotWidget.remove_linear_region()
        self.dock_widget.setVisible(False)
        self.accepted_func()
        # super(FitWidget, self).accept()

    def reject(self):
        # self.remove_last_fit()
        self.clear_plot()
        FitWidget.is_opened = False
        FitWidget._instance = None
        # self.plot_widget.removeItem(self.lr)
        # del self.lr
        PlotWidget.remove_linear_region()
        self.dock_widget.setVisible(False)

        # super(FitWidget, self).reject()


if __name__ == "__main__":
    import sys

    def my_exception_hook(exctype, value, traceback):
        # Print the error and traceback
        print(exctype, value, traceback)
        # Call the normal Exception hook after
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)


    sys._excepthook = sys.excepthook

    # Set the exception hook to our wrapping function
    sys.excepthook = my_exception_hook

    app = QtWidgets.QApplication(sys.argv)
    Dialog = FitWidget(None, None)
    # Dialog.show()
    sys.exit(app.exec_())



import sys
import os
from PyQt5 import QtCore
from spectramanipulator import windows

# from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt, QCoreApplication
from PyQt5.QtWidgets import QMessageBox, QMainWindow, QFileDialog, QWidget, QPushButton, QStatusBar, QLabel, \
    QDockWidget, QApplication

from PyQt5.QtGui import QColor, QFont
from PyQt5 import QtWidgets
import argparse
from spectramanipulator import __version__

import pyqtgraph as pg

from .treewidget import TreeWidget, get_hierarchic_list
from .treeview.model import ItemIterator
from .project import Project
from .settings import Settings
from .logger import Logger, Transcript
from .plotwidget import PlotWidget

from .user_namespace import UserNamespace
from .menubar import MenuBar
from .console import Console
from .misc import intColor, intColorGradient, int_default_color_scheme

from .dialogs.settingsdialog import SettingsDialog
from .treeview.item import SpectrumItemGroup
from .dataloader import parse_files_specific
from .spectrum import SpectrumList
from .dialogs.fitwidget import FitWidget
from .dialogs.load_kinetics_dialog import LoadKineticsDialog
import re

from .associate_project_file import associate_project_file, remove_project_file_association

import numpy as np


debug = False

import cProfile
import pstats


class Main(QMainWindow):
    current_file = None

    def __init__(self, filepath=None, parent=None):
        super(Main, self).__init__(parent)

        self.setWindowTitle("Untitled - Simple Spectra Manipulator")

        w, h = 1000, 600

        # setup window size based on current resolution
        if sys.platform == 'win32':
            from win32api import GetSystemMetrics
            w, h = int(GetSystemMetrics(0) * 0.45), int(GetSystemMetrics(1) * 0.45)

        # print(w, h)
        self.resize(w, h)

        self.console = Console(self)
        sys.stdout = Transcript()

        self.dockTreeWidget = QDockWidget(self)
        self.dockTreeWidget.setTitleBarWidget(QWidget())
        self.dockTreeWidget.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.dockTreeWidget.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea)

        # create and set variable docking widget, where different function widows will appear
        self.var_widget = QDockWidget(self)
        lbl = QLabel('Some Title Text')
        lbl.setAlignment(Qt.AlignCenter)
        font = QFont()
        font.setBold(True)
        lbl.setFont(font)
        self.var_widget.setTitleBarWidget(lbl)
        self.var_widget.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.var_widget.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea)
        self.var_widget.setVisible(False)
        # self.var_widget.titleBarWidget()

        self.tree_widget = TreeWidget(self.dockTreeWidget)
        self.dockTreeWidget.setWidget(self.tree_widget)

        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockTreeWidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.var_widget)
        self.addDockWidget(Qt.BottomDockWidgetArea, self.console)

        # fixing the resize bug https://stackoverflow.com/questions/48119969/qdockwidget-splitter-jumps-when-qmainwindow-resized
        self.resizeDocks([self.dockTreeWidget], [3 * w // 7], Qt.Horizontal)
        # self.resizeDocks([self.var_widget], [250], Qt.Vertical)
        self.resizeDocks([self.console], [h // 3], Qt.Vertical)
        self.setCorner(Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)

        Settings.load()

        self.coor_label = QLabel()
        self.grpView = PlotWidget(self, coordinates_func=self.coor_label.setText)
        self.setCentralWidget(self.grpView)

        self.createStatusBar()
        self.logger = Logger(self.console.showMessage, self.statusBar().showMessage)

        self.tree_widget.redraw_spectra.connect(self.redraw_all_spectra)
        self.tree_widget.state_changed.connect(self.add_star)

        self.user_namespace = UserNamespace(self)

        self.setMenuBar(MenuBar(self))

        self.update_recent_files()

        # open project or file if opened with an argument
        if filepath is not None:
            self.open_project(filepath=filepath, open_dialog=False)

    def add_star(self):
        if not self.windowTitle().startswith('*'):
            self.setWindowTitle('*' + self.windowTitle())

    def createStatusBar(self):
        statusBar = QStatusBar()
        self.setStatusBar(statusBar)
        statusBar.showMessage("Ready", 3000)
        self.console_button = QPushButton("Console")
        self.console_button.setFlat(True)
        self.console_button.setCheckable(True)
        self.console_button.toggled.connect(self.console.setVisible)
        statusBar.addPermanentWidget(self.coor_label)

        statusBar.addPermanentWidget(self.console_button)
        self.console_button.isChecked()

    def update_current_file(self, filepath):
        self.current_file = filepath
        head, tail = os.path.split(filepath)
        self.setWindowTitle(os.path.splitext(tail)[0] + ' - Spectra Manipulator')

    def update_recent_files(self):
        # num = min(len(Settings.recent_project_filepaths), len(self.menuBar().recent_file_actions))
        num = len(Settings.recent_project_filepaths)

        # update all of them
        for i in range(num):
            filepath = Settings.recent_project_filepaths[i]
            head, tail = os.path.split(filepath)
            text = os.path.split(head)[1] + '/' + os.path.splitext(tail)[0]
            self.menuBar().recent_file_actions[i].setText(text)
            self.menuBar().recent_file_actions[i].setData(filepath)  # save filepath as data to QAction
            self.menuBar().recent_file_actions[i].setVisible(True)
            self.menuBar().recent_file_actions[i].setStatusTip(filepath)  # show filepath in statusbar when hovered

        for i in range(num, self.menuBar().MAX_RECENT_FILES):  # set invisible the rest
            self.menuBar().recent_file_actions[i].setVisible(False)

    def add_recent_file(self, filepath):
        filepath = filepath.replace('\\', '/')  # replace \ for /

        # if there is the same filepath in the list, remove this entry
        if filepath in Settings.recent_project_filepaths:
            Settings.recent_project_filepaths.remove(filepath)

        Settings.recent_project_filepaths.insert(0, filepath)
        Settings.recent_project_filepaths = Settings.recent_project_filepaths[:self.menuBar().MAX_RECENT_FILES]

        Settings.save()

        self.update_recent_files()

    def actioncopy_to_svg_clicked(self):

        self.grpView.save_plot_to_clipboard_as_svg()

    # on close
    def closeEvent(self, event):
        if self.windowTitle().startswith('*') and self.tree_widget.top_level_items_count() > 0:
            reply = QMessageBox.question(self, 'Message', "Do you want to save the project?", QMessageBox.Yes |
                                         QMessageBox.No | QMessageBox.Cancel)

            if reply == QMessageBox.Yes:
                self.save_project()
                Settings.save()
            if reply == QMessageBox.No:
                event.accept()
            else:
                event.ignore()

    def open_settings(self):

        if SettingsDialog.is_opened:
            SettingsDialog.get_instance().activateWindow()
            SettingsDialog.get_instance().setFocus()
            return

        sett_dialog = SettingsDialog()

        if not sett_dialog.accepted:
            return

        self.grpView.update_settings()
        self.redraw_all_spectra()

    def batch_load_kinetics(self):
        """Opens a Batch Load Kinetics dialog and then call the function from treewidget"""
        if LoadKineticsDialog.is_opened:
            LoadKineticsDialog.get_instance().activateWindow()
            LoadKineticsDialog.get_instance().setFocus()
            return

        load_kin_dialog = LoadKineticsDialog()

        if not load_kin_dialog.accepted:
            return

        paths = load_kin_dialog.lwFolders.paths
        if len(paths) == 0:
            return

        try:
            spectra_dir_name = load_kin_dialog.leSpectra.text()
            times_name = load_kin_dialog.leTimes.text()
            blank_name = load_kin_dialog.leBlank.text()
            dt = float(load_kin_dialog.leTimeUnit.text()) if load_kin_dialog.cbKineticsMeasuredByEach.isChecked() else None
            bcorr_range = (float(load_kin_dialog.leBCorr0.text()), float(load_kin_dialog.leBCorr1.text())) \
                if load_kin_dialog.cbBCorr.isChecked() else None

            cut_range = (float(load_kin_dialog.leCut0.text()), float(load_kin_dialog.leCut1.text())) \
                if load_kin_dialog.cbCut.isChecked() else None

            for path in paths:
                self.tree_widget.load_kinetic(path, spectra_dir_name, times_name, blank_name,
                                              dt=dt, b_corr=bcorr_range, cut=cut_range)
        except Exception as ex:
            Logger.message(f"Unable to load kinetics: {ex.__str__()}")

    @staticmethod
    def _open_file_dialog(caption='Open ...', initial_dir='...', _filter='All Files (*.*)',
                          initial_filter='All Files (*.*)', choose_multiple=False):

        f = QFileDialog.getOpenFileNames if choose_multiple else QFileDialog.getOpenFileName


        filepaths = f(caption=caption,
                      directory=initial_dir,
                      filter=_filter,
                      initialFilter=initial_filter)

        if not choose_multiple and filepaths[0] == '':
            return None

        if choose_multiple and len(filepaths[0]) < 1:
            return None

        return filepaths[0]

    def open_project(self, filepath=None, open_dialog=True):

        if open_dialog:
            filepath = self._open_file_dialog("Open project", Settings.open_project_dialog_path,
                                              _filter=f"Project files (*{Settings.PROJECT_EXTENSION});;All Files (*.*)",
                                              initial_filter=f"Project files (*{Settings.PROJECT_EXTENSION})",
                                              choose_multiple=False)

            if filepath is None:
                return

            Settings.open_project_dialog_path = os.path.dirname(filepath)
            Settings.save()

        if not os.path.exists(filepath):
            filepath = filepath.replace('\\', '/')
            Logger.message(f"File {filepath} does not exist.")
            if filepath in Settings.recent_project_filepaths:
                Settings.recent_project_filepaths.remove(filepath)
                Settings.save()

            self.update_recent_files()
            return

        project = Project.deserialize(filepath)
        # try:
        # except Exception as err:
        #     Logger.message("Unable to load {}.\n{}".format(filepath, err.__str__()))
        #     return

        if self.tree_widget.top_level_items_count() != 0:
            reply = QMessageBox.question(self, 'Open project', "Do you want to merge the project with current project? "
                                                               "By clicking No, current project will be deleted and replaced.",
                                         QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            if reply == QMessageBox.Yes:
                pass
            elif reply == QMessageBox.No:
                # delete all spectra and import new
                self.tree_widget.clear()
                project.settings.set_settings()
            else:
                return

        if not hasattr(project, 'spectra_list') or project.spectra_list is None:  # for backward compatibility
            self.tree_widget.import_project(project.generic_item)
        else:  # for backward compatibility, for old project structure
            self.tree_widget.import_spectra(project.spectra_list)
        self.add_recent_file(filepath)

        self.update_current_file(filepath)

    def save_project_as(self):

        if self.tree_widget.top_level_items_count() == 0:
            return

        _filter = f"Project files (*{Settings.PROJECT_EXTENSION})"

        filepath = QFileDialog.getSaveFileName(caption="Save project",
                                               directory=Settings.save_project_dialog_path if self.current_file is None else self.current_file,
                                               filter=_filter, initialFilter=_filter)

        if filepath[0] == '':
            return

        Logger.message(f"Saving project to {filepath[0]}")

        Settings.save_project_dialog_path = os.path.dirname(filepath[0])
        Settings.save()

        project = Project(generic_item=self.tree_widget.myModel.root)
        project.serialize(filepath[0])

        self.update_current_file(filepath[0])
        self.add_recent_file(filepath[0])
        Logger.message("Done")

    def save_project(self):

        if self.tree_widget.top_level_items_count() == 0:
            return

        if self.current_file is not None and os.path.exists(self.current_file):
            project = Project(generic_item=self.tree_widget.myModel.root)
            project.serialize(self.current_file)

            Logger.status_message(f"Project was saved to {self.current_file}.")

            self.update_current_file(self.current_file)
        else:
            self.save_project_as()

    def file_menu_import_files(self):
        filepaths = self._open_file_dialog("Import files", Settings.import_files_dialog_path,
                                           _filter="Data Files (*.txt, *.TXT, *.csv, *.CSV, *.dx, *.DX);;All Files (*.*)",
                                           initial_filter=f"All Files (*.*)",
                                           choose_multiple=True)

        if filepaths is None:
            return

        Settings.import_files_dialog_path = os.path.dirname(filepaths[0])
        Settings.save()

        self.tree_widget.import_files(filepaths)

    def import_LPF_kinetics(self):
        filepaths = self._open_file_dialog("Import LFP Kinetics", Settings.import_LPF_dialog_path,
                                           _filter="Data Files (*.csv, *.CSV);;All Files (*.*)",
                                           initial_filter="Data Files (*.csv, *.CSV)",
                                           choose_multiple=True)

        if filepaths is None:
            return

        Settings.import_LPF_dialog_path = os.path.dirname(filepaths[0])
        Settings.save()

        kwargs = dict(delimiter=',',
                      decimal_sep='.',
                      remove_empty_entries=False,
                      skip_col_num=3,
                      general_import_spectra_name_from_filename=True,
                      skip_nan_columns=False,
                      nan_replacement=0)

        spectra, _ = parse_files_specific(filepaths, use_CSV_parser=False, **kwargs)
        if len(spectra) == 0:
            return

        max_abs = 3.5

        # CONVERT THE VOLTAGE (proportional to transmittance) TO ABSORBANCE
        for sp in spectra:
            # convert nan and pos infinities to value of max_abs = very high absorbance for nano
            sp.y = np.nan_to_num(-np.log10(-sp.y), nan=max_abs, posinf=max_abs, neginf=0)
            sp.y[sp.y > max_abs] = max_abs  # floor larger values to max absorbance

        self.tree_widget.import_spectra(spectra)

    def import_EEM_Duetta(self):
        """Used to import excitation emission map from Duetta Fluorimeter.
        Works only for proper data. Data has to be exported from finished kinetics. If the kinetic
        measurement was stopped during the measurement, the exported data will have different
        format and they cannot be imported this way.

        It sets extracted excitation wavelengths as the names of the spectra.
        """

        filepaths = self._open_file_dialog("Import Excitation Emission Map from Duetta Fluorimeter",
                                           Settings.import_EEM_dialog_path,
                                           _filter="Data Files (*.txt, *.TXT);;All Files (*.*)",
                                           initial_filter="Data Files (*.txt, *.TXT)",
                                           choose_multiple=True)

        if filepaths is None:
            return

        Settings.import_EEM_dialog_path = os.path.dirname(filepaths[0])
        Settings.save()

        kwargs = dict(delimiter='\t',
                      decimal_sep='.',
                      remove_empty_entries=False,
                      skip_col_num=0,
                      general_import_spectra_name_from_filename=True,
                      skip_nan_columns=False,
                      nan_replacement=0)

        spectra, parsers = parse_files_specific(filepaths, use_CSV_parser=False, **kwargs)
        if len(spectra) == 0:
            return

        if not isinstance(spectra[0], SpectrumList):
            Logger.message(f"{type(spectra[0])} is not type SpectrumList. "
                           f"Unable to import data. Check the dataparsers.")
            return

        # eg. MeCN blank 2nm step 448:250-1100, [name] [ex]:[em1]-[em2]
        # we need 448 which is the excitation wavelength
        pattern = re.compile(r'(\d+):\d+-\d+')  # use regex to extract the ex. wavelength

        try:
            # extract the wavelengths from parsers
            for sl, parser in zip(spectra, parsers):
                names_list = parser.names_history[0]  # first line in names history
                assert len(names_list) == len(sl) + 1

                # extract the excitation wavelengths from the name history
                new_names = []
                sl_name = None
                for name in names_list:
                    if name == '':
                        continue
                    m = pattern.search(name)
                    if m is None:
                        continue
                    new_names.append(m.group(1))
                    sl_name = name.replace(m.group(0), '').strip()

                # set the main name
                if sl_name:
                    sl.name = sl_name

                # remove each 2nd spectrum as it contains useless X values (starting from second spectrum)
                del sl.children[1::2]

                # setup extracted names = excitation wavelengths
                sl.set_names(new_names)

                # 'sort' the list, the data are imported in opposite way so we can just reverse the list
                sl.children = sl.children[::-1]
        except Exception as e:
            Logger.message(f"Unable to import data: {e.__str__()}")
            return

        self.tree_widget.import_spectra(spectra)

    def import_kinetics_Duetta(self):
        """Used to import emission kinetics from Duetta Fluorimeter.
        Works only for proper data. Data has to be exported from finished kinetics. If the kinetic
        measurement was stopped during the measurement, the exported data will have different
        format and they cannot be imported this way.

        It sets extracted times as a names of the spectra. First spectrum will have time=0.
        In other words, time of first spectrum will be subtracted from all spectra.
        """

        filepaths = self._open_file_dialog("Import Kinetics from Duetta Fluorimeter",
                                           Settings.import_EEM_dialog_path,
                                           _filter="Data Files (*.txt, *.TXT);;All Files (*.*)",
                                           initial_filter="Data Files (*.txt, *.TXT)",
                                           choose_multiple=True)

        if filepaths is None:
            return

        Settings.import_EEM_dialog_path = os.path.dirname(filepaths[0])
        Settings.save()

        kwargs = dict(delimiter='\t',
                      decimal_sep='.',
                      remove_empty_entries=False,
                      skip_col_num=0,
                      general_import_spectra_name_from_filename=True,
                      skip_nan_columns=False,
                      nan_replacement=0)

        spectra, parsers = parse_files_specific(filepaths, use_CSV_parser=False, **kwargs)
        if len(spectra) == 0:
            return

        if not isinstance(spectra[0], SpectrumList):
            Logger.message(f"{type(spectra[0])} is not type SpectrumList. "
                           f"Unable to import data. Check the dataparsers.")
            return

        # eg. 2Z+MB 310:250-650,1.92,  [name] [ex]:[em1]-[em2],[time]
        # we need 1.92 which is the time at it was measured
        pattern = re.compile(r'\d+:\d+-\d+,(\d+.\d+)')  # matches the [ex]:[em1]-[em2],[time] pattern and use [time] as group

        try:
            # extract the times from parsers
            for sl, parser in zip(spectra, parsers):
                names_list = parser.names_history[0]  # first line in names history
                assert len(names_list) == len(sl) + 1

                # extract the excitation wavelengths from the name history
                new_names = []
                sl_name = None  # name of the group
                first_time = None
                for name in names_list:
                    if name == '':
                        continue
                    m = pattern.search(name)
                    if m is None:
                        continue

                    time = float(m.group(1))
                    if first_time is None:
                        first_time = time

                    new_names.append(f'{time - first_time:.2f}')  # use 2 digits precision
                    sl_name = name.replace(m.group(0), '').strip()

                # set the main name
                if sl_name:
                    sl.name = sl_name

                # remove each 2nd spectrum as it contains useless X values (starting from second spectrum)
                del sl.children[1::2]

                # setup extracted names = excitation wavelengths
                sl.set_names(new_names)
        except Exception as e:
            Logger.message(f"Unable to import data: {e.__str__()}")
            return

        self.tree_widget.import_spectra(spectra)

    def intLineStyle(self, counter):
        styles = [Qt.SolidLine, Qt.DashLine, Qt.DotLine, Qt.DashDotLine, Qt.DashDotDotLine]
        return styles[counter % len(styles)]

    def get_user_gradient(self):
        """Gradient in a format of
        position (0, 1) \t R \t G \t B \t A \n
        etc.

        Entries separated by tabulator \t and lines by new line \n

        """
        try:
            lines = Settings.user_defined_grad.split('\n')
            lines = list(filter(None, lines))  # remove empty entries

            data = np.zeros((len(lines), 5), dtype=np.float32)

            for i, line in enumerate(lines):  # parse the string data into matrix
                entries = line.split('\t')
                data[i] = np.asarray([float(entry) for entry in entries])

            data[:, 1:] *= 255  # multiply the rgba values by 255

            return data
        except:
            pass
            # Console.showMessage("User defined color scheme is not correct.")

    def redraw_all_spectra(self):
        self.grpView.clear_plots()
        self.redraw_items()

    def update_items_data(self, items):
        self.grpView.update_items_data(items)

    def redraw_items(self, items=None, remove=None):

        if remove is True:
            for item in items:
                self.grpView.remove(item)

        gradient_mat = None
        if Settings.color_scheme == 2:  # user defined
            gradient_mat = self.get_user_gradient()
            if gradient_mat is None:
                Console.showMessage("Cannot plot the spectra, user defined gradient matrix is not correct.")
                return

        # item_counter = 0
        # iterate over all checked spectra items and draw them
        # it = QTreeWidgetItemIterator(self.treeWidget, QTreeWidgetItemIterator.Checked)

        # for item in self.tree_widget.myModel.iterate_items(ItemIterator.Checked):
        # iterator = items if items is not None else self.tree_widget.myModel.iterate_items(ItemIterator.Checked)

        group_counter = -1
        last_group = None

        # iterate over all checked items, if iterated item is already plotted, continue
        for item_counter, item in enumerate(self.tree_widget.myModel.iterate_items(ItemIterator.Checked)):

            if isinstance(item, SpectrumItemGroup):
                # last_group = item
                continue

            # check if we are plotting a new group, if so, increment counter
            if item.is_in_group():
                if last_group != item.parent:
                    last_group = item.parent
                    group_counter += 1

            # if item in self.grpView.plotted_items:  # look for keys in dictionary
            #     continue

            # if the spectra item is part of a group, write the group name in square brackets before the spectra name in legend
            spectrum_name = '<strong>{}</strong>: {}'.format(item.parent.name, item.name) if item.is_in_group() \
                else item.name

            style = self.intLineStyle(
                group_counter) if Settings.different_line_style_among_groups and item.is_in_group() else Qt.SolidLine

            counter = group_counter if Settings.same_color_in_group and item.is_in_group() else item_counter

            if Settings.color_scheme == 0:
                color = int_default_color_scheme(counter)
            elif Settings.color_scheme == 1:
                color = intColor(counter,
                                 hues=Settings.hues,
                                 values=Settings.values,
                                 maxValue=Settings.maxValue,
                                 minValue=Settings.minValue,
                                 maxHue=Settings.maxHue,
                                 minHue=Settings.minHue,
                                 sat=Settings.sat,
                                 alpha=Settings.alpha,
                                 reversed=Settings.HSV_reversed)
            else:
                color = intColorGradient(counter, Settings.hues, gradient_mat, reversed=Settings.HSV_reversed)

            try:
                line_alpha = item.line_alpha if hasattr(item, 'line_alpha') else 255
                line_color = color if item.color is None else QColor(*item.color) if isinstance(item.color, (
                    tuple, list)) else QColor(item.color)  # if string - html format or name of color
                if Settings.color_scheme < 2:  # only for default and HSV
                    line_color.setAlpha(line_alpha)
                pen = pg.mkPen(color=line_color,
                               width=Settings.line_width if item.line_width is None else item.line_width,
                               style=style if item.line_type is None else item.line_type)
            except AttributeError:
                pen = pg.mkPen(color=color, width=Settings.line_width, style=style)

            try:
                symbol = item.symbol

                _brush = item.symbol_brush if item.symbol_brush is not None else pen.color().name()
                symbolPen = QColor(*_brush) if isinstance(_brush, (tuple, list)) else QColor(_brush)
                symbolPen.setAlpha(item.sym_brush_alpha)
                _fill = item.symbol_fill if item.symbol_fill is not None else pen.color().name()
                symbolBrush = QColor(*_fill) if isinstance(_fill, (tuple, list)) else QColor(_fill)
                symbolBrush.setAlpha(item.sym_fill_alpha)
                symbol_size = item.symbol_size
            except AttributeError:
                symbol = None
                symbolBrush = None
                symbolPen = None
                symbol_size = None

            self.grpView.plot(item,
                              pen=pen,
                              name=spectrum_name,
                              plot_legend=item.plot_legend if hasattr(item, 'plot_legend') else True,
                              symbolBrush=symbolBrush,
                              symbolPen=symbolPen,
                              symbol=symbol,
                              symbolSize=symbol_size,
                              zValue=item_counter if Settings.reverse_z_order else -item_counter)

        FitWidget.replot()  # replot all fits if FitWidget is active


def my_exception_hook(exctype, value, traceback):
    # Print the error and traceback
    print(exctype, value, traceback)
    # Call the normal Exception hook after
    sys._excepthook(exctype, value, traceback)
    sys.exit(1)


def main():
    global debug
    # Back up the reference to the exceptionhook
    sys._excepthook = sys.excepthook

    # Set the exception hook to our wrapping function
    sys.excepthook = my_exception_hook

    parser = argparse.ArgumentParser('spectra-manipulator')
    parser.add_argument("--debug", action="store_true",
                        help="runs in debug mode (keeps the console visible)")
    parser.add_argument("--version", action="version",
                        version="%(prog)s {}".format(__version__))
    parser.add_argument("--associate_files", action="store_true",
                        help="associates project files for windows and exits")
    parser.add_argument("--remove_association", action="store_true",
                        help="removes project files association from windows registry and exits")

    args, unparsed_args = parser.parse_known_args()
    debug = args.debug

    if args.associate_files:
        associate_project_file()
        return

    if args.remove_association:
        remove_project_file_association()
        return

    try:
        if sys.platform == 'win32' and not debug:
            windows.set_attached_console_visible(False)

        app = QApplication(sys.argv)

        QCoreApplication.setOrganizationName("Spectra Manipulator")
        QCoreApplication.setOrganizationDomain("FreeTimeCoding")
        QCoreApplication.setApplicationName("Spectra Manipulator")

        # app.setStyle('Windows')
        # pass the filename if provided
        form = Main(unparsed_args[0] if len(unparsed_args) > 0 else None)
        form.show()

        # form.interact()
        # app.lastWindowClosed.connect(app.quit)
        sys.exit(app.exec_())
    finally:
        if sys.platform == 'win32' and not debug:
            windows.set_attached_console_visible(True)

    ### cProfile.run('app.exec_()', 'profdata')
    # cProfile.runctx('app.exec_()', None, locals(), filename='profdata')
    # p = pstats.Stats('profdata')
    # p.sort_stats('cumtime').print_stats(100)


if __name__ == "__main__":
    main()

# print(QtWidgets.QStyleFactory.keys())



from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QApplication

import pyqtgraph.opengl as gl
import pyqtgraph as pg
import numpy as np
from .spectrum import group2mat
import sys


class GLPlotter:
    def __init__(self, group_item=None):

        self.x, self.y, self.mat = group2mat(group_item)
        if self.y is None:
            return

        self.x -= self.x.mean()
        self.y -= self.y.mean()
        self.mat *= 50

        self.traces = []
        self.w = gl.GLViewWidget()
        self.w.setCameraPosition(distance=500)
        self.w.setWindowTitle(f'Plot of {group_item.name}')
        self.w.setGeometry(0, 110, 1400, 700)
        self.w.show()

        gax = gl.GLAxisItem(size=QtGui.QVector3D(self.x.mean(), self.y.mean(), self.mat.min()))
        self.w.addItem(gax)

        self.plot_data()

    def plot_data(self):
        # for i in range(self.y.shape[0]):
        #     pts = np.vstack((self.x, np.ones_like(self.x) * self.y[i], self.mat[:, i])).T
        #     self.traces.append(gl.GLLinePlotItem(pos=pts, color=pg.glColor(
        #         (i, self.y.shape[0] * 1.3)), width=2, antialias=True))
        #     self.w.addItem(self.traces[i])

        surface = gl.GLSurfacePlotItem(x=self.x, y=self.y, z=self.mat, shader='heightColor', drawEdges=True)
        surface.shader()['colorMap'] = np.array([0.2, 2, 0.5])
        self.w.addItem(surface)

    #
    # def set_plotdata(self, name, points, color, width):
    #     self.traces[name].setData(pos=points, color=color, width=width)
    #
    # def update(self):
    #     for i in range(self.n):
    #         yi = np.array([self.y[i]] * self.m)
    #         d = np.sqrt(self.x ** 2 + yi ** 2)
    #         z = 10 * np.cos(d + self.phase) / (d + 1)
    #         pts = np.vstack([self.x, yi, z]).transpose()
    #         self.set_plotdata(
    #             name=i, points=pts,
    #             color=pg.glColor((255, 0, 0, 255*i/self.n)),
    #             width=(i + 1) / 10
    #         )
    #         self.phase -= .001


# Start Qt event loop unless running in interactive mode.
if __name__ == '__main__':
    v = GLPlotter()



import csv
from spectramanipulator.parsers.generalparser import GeneralParser


class CSVFileParser(GeneralParser):

    def __init__(self, filepath=None, str_data=None, delimiter=',', decimal_sep='.', remove_empty_entries=True,
                 skip_col_num=0, general_import_spectra_name_from_filename=False,
                 general_if_header_is_empty_use_filename=True, doublequote=True, skipinitialspace=True,
                 skip_nan_columns=False, nan_replacement=0):
        super(CSVFileParser, self).__init__(filepath, str_data, delimiter, decimal_sep, remove_empty_entries,
                                            skip_col_num,
                                            general_import_spectra_name_from_filename,
                                            general_if_header_is_empty_use_filename,
                                            skip_nan_columns, nan_replacement)

        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace

    def line2list_iterator(self):
        # get line iterator, return parsed list iterator, like eg. ['190', '1.5', '1.3', 'some value']
        it = self.__iter__()
        return csv.reader(it, doublequote=self.doublequote,
                          skipinitialspace=self.skipinitialspace,
                          delimiter=self.delimiter if self.delimiter != '' else ',')



from abc import abstractmethod
import sys
from sys import platform
from copy import deepcopy

from PyQt5.QtWidgets import QTreeView, QAbstractItemView, QStyledItemDelegate
from PyQt5.QtCore import pyqtSignal, QAbstractItemModel, QVariant, QModelIndex, QItemSelection, QItemSelectionModel, Qt
from PyQt5.QtGui import QFont

from .pymimedata import PyMimeData
from .item import GenericItem, SpectrumItemGroup, SpectrumItem

# from spectramanipulator import Spectrum, SpectrumList

import cProfile
import pstats

import numpy as np

# possible workaround - https://stackoverflow.com/questions/34419072/how-to-improve-selection-performance-with-pyside-qtcore-qabstractitemmodel-and-q

# this was tough to get together, I could not use the TreeWidget with StandardItemModel,
# because iteration over items sometime gave not my item (SpectrumItem or SpectrumItemGroup) with my data,
# but the blank TreeWidgetItem, which raised errors and that sucked a lot, therefore, only solution was to
# implement QAbstractItemModel and then use TreeView instead of TreeWidget
# at least, now, it really works, but I had to write everything by myself... great thing about it is moving the
# items with data in between two separate opened applications, it would be almost impossible to do that in
# QTreeWidget...
# eventually, it is slower than QTreeWidget, but it is just python and Qt, I cannot expect miracles from it

# I got help from many sources...

# https://stackoverflow.com/questions/841096/slow-selection-in-qtreeview-why
# https://stackoverflow.com/questions/8175122/qtreeview-checkboxes
# https://riverbankcomputing.com/pipermail/pyqt/2009-April/022729.html
# http://flame-blaze.net/archives/5249
# http://doc.qt.io/archives/qt-4.8/qstandarditemmodel.html


class ItemIterator:
    NoChildren = 0
    Checked = 1
    All = 2
    Groups = 3


class Model(QAbstractItemModel):
    checked_changed_signal = pyqtSignal(list, bool)
    data_dropped_signal = pyqtSignal()
    item_edited_signal = pyqtSignal(bool)
    all_unchecked_signal = pyqtSignal()
    data_modified_signal = pyqtSignal(list)  # list of spectra that were modified
    info_modified_signal = pyqtSignal(list)  # list of spectra that were modified
    items_sorted_signal = pyqtSignal(list)
    items_ungrouped_signal = pyqtSignal(list)
    items_moved_signal = pyqtSignal(list)

    def __init__(self, parent=None):
        super(Model, self).__init__(parent)

        self.treeView = parent
        self.headers = ['Name', 'Index | Info']

        self.columns = 2

        self.coarsed_mimeData = None

        GenericItem.data_modified_signal = self.data_modified_signal
        GenericItem.info_modified_signal = self.info_modified_signal

        # Create items
        self.root = GenericItem('root', 'this is root', None)

        # itemA = SpectrumItemGroup('itemA', 'this is item A', self.root)
        # itemA1 = SpectrumItem(None, 'itemA1', 'this is item A1', itemA)
        # #
        # itemB = SpectrumItemGroup('itemB', 'this is item B', self.root)
        # itemB1 = SpectrumItem(None, '1', 'this is item B1', itemB)
        # itemB2 = SpectrumItem(None, '2', 'this is item B1', itemB)
        # itemB3 = SpectrumItem(None, '3', 'this is item B1', itemB)
        #
        # itemC = SpectrumItemGroup('itemC', 'this is item C', self.root)
        # itemC1 = SpectrumItem(None, 'itemC1', 'this is item C1', self.root)

    def iterate_items(self, what_items):
        """
        Iterate specified items of TreeView.

        :param what_items: ItemIterator
        """
        if what_items == ItemIterator.Groups:
            for item in self.root.children:
                if isinstance(item, SpectrumItemGroup):
                    yield item
        elif what_items == ItemIterator.All:
            for item in self.root.children:
                yield item
                for child in item.children:
                    yield child
        elif what_items == ItemIterator.Checked:
            for item in self.root.children:
                if item.__len__() == 0:
                    if item.check_state == Qt.Checked or item.check_state == Qt.PartiallyChecked:
                        yield item
                        continue
                for child in item.children:
                    if child.check_state == Qt.Checked:
                        yield child
        elif what_items == ItemIterator.NoChildren:
            for item in self.root.children:
                if item.__len__() == 0:
                    yield item
                for child in item.children:
                    if child.__len__() == 0:
                        yield child

    def iterate_selected_items(self, skip_groups=False, skip_childs_in_selected_groups=True,
                               clear_selection=False):

        indexes = self.treeView.selectedIndexes()

        if len(indexes) == 0:
            return

        selected_items = []

        # step is two, because we have two columns, so there is 2 indexes per one item
        # first we have to sort the selected items because the are not sorted...
        for i in range(0, len(indexes), 2):
            idx = indexes[i]
            item = self.node_from_index(indexes[i])

            order = int(100000000)
            if item.is_top_level():
                # ix = self.indexFromItem(item)
                # row = idx.row()
                # print(row)
                order *= (idx.row() + 1)
            else:
                order *= (self.root.rowOfChild(item.parent) + 1)
                order += idx.row() + 1
            selected_items.append((item, order))

        selected_items = sorted(selected_items, key=lambda x: x[1])
        sorted_sel_items = []
        for item in selected_items:
            sorted_sel_items.append(item[0])

        if clear_selection:
            self.treeView.clearSelection()

        last_group = None
        for item in sorted_sel_items:
            if isinstance(item, SpectrumItemGroup):
                last_group = item
                if not skip_groups:
                    yield item
            if isinstance(item, SpectrumItem):
                # skip the child if only iterating over childs, whose parents are not selected
                if skip_childs_in_selected_groups and item.parent == last_group:
                    continue
                yield item

    def supportedDropActions(self):
        return Qt.MoveAction | Qt.CopyAction

    def flags(self, index):
        defaultFlags = QAbstractItemModel.flags(self, index)

        if index.isValid():

            if index.column() == 0:
                return Qt.ItemIsEditable | Qt.ItemIsUserCheckable | Qt.ItemIsTristate | Qt.ItemIsDragEnabled | \
                       Qt.ItemIsDropEnabled | defaultFlags
            else:
                return Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled | defaultFlags

        else:
            return Qt.ItemIsDropEnabled | defaultFlags

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self.headers[section])
        return QVariant()

    def mimeTypes(self):
        return [PyMimeData.MIME_TYPE]

    def mimeData(self, Iterable, q_model_index=None):

        item_list = [item for item in self.iterate_selected_items(False, True)]

        mimeData = PyMimeData(item_list)
        return mimeData

    def canDropMimeData(self, mimedata, action, row, column, parentIndex):
        parentNode = self.node_from_index(parentIndex)
        # print(action, row, column, parentNode.name)
        if self.coarsed_mimeData is None and not isinstance(mimedata, PyMimeData):

            if mimedata.hasUrls():
                for url in mimedata.urls():
                    if url.isLocalFile():
                        return True
                return False
            if mimedata.hasText():
                return True
            if mimedata.hasFormat("XML Spreadsheet"):
                return True
            if mimedata.hasFormat(PyMimeData.MIME_TYPE):
                mimedata = PyMimeData.coerce(mimedata)
                self.coarsed_mimeData = mimedata
            return False

        if not isinstance(mimedata, PyMimeData) and self.coarsed_mimeData is not None:
            mimedata = self.coarsed_mimeData
        # else:
        #     raise Exception("canDropMimeData - MimeData is not type of PyMimeData")

        dropped_data = mimedata.instance()

        if isinstance(parentNode, SpectrumItem):
            return False

        if parentNode.is_root():
            return True

        contain_group = False
        for item in dropped_data:
            if isinstance(item, SpectrumItemGroup):
                contain_group = True
                break

        if contain_group:
            return False

        return True

    def dropMimeData(self, mimedata, action, row, column, parentIndex):
        if not isinstance(mimedata, PyMimeData):
            mimedata = PyMimeData.coerce(mimedata)

        if action == Qt.IgnoreAction:
            return True

        dragNodes = mimedata.instance()
        parentNode = self.node_from_index(parentIndex)

        row_to_insert = len(parentNode) if row == -1 else row

        # make an copy of the node being moved
        newNodes = deepcopy(dragNodes)
        for node in reversed(newNodes):
            node.setParent(parentNode, row=row_to_insert)

        self.insertRows(row_to_insert, len(dragNodes), parentIndex)

        self._update_tristate(parentNode)
        self.dataChanged.emit(parentIndex, parentIndex)
        self.data_dropped_signal.emit()
        # self.emit(SIGNAL("dataChanged(QModelIndex,QModelIndex)"),
        #           parentIndex, parentIndex)

        self.coarsed_mimeData = None
        return True

    def take_all_children(self, group_item):
        parent_index = self.index_from_node(group_item)
        self.beginRemoveRows(parent_index, 0, len(group_item.children) - 1)
        ref_children_list = group_item.children
        group_item.children = []  # reassign to empty list
        self.endRemoveRows()
        return ref_children_list

    def take_item(self, item):
        parent_index = self.index_from_node(item.parent)
        row = item.parent.rowOfChild(item)
        self.beginRemoveRows(parent_index, row, row)
        item.parent.removeChildAtRow(row)
        self.endRemoveRows()
        # item.setParent(None)
        return item

    # def move_items(self, items):
    #
    #     source_parent = self.myModel.index_from_node(last_parent)
    #
    #     self.myModel.beginMoveRows(source_parent, first_row, last_row, group_item_index, i)
    #
    #     chunk = last_parent.children[first_row:last_row + 1]
    #     del last_parent.children[first_row:last_row + 1]
    #
    #     for child in chunk:
    #         child.setParent(group_item)
    #
    #     # item.parent.removeChildAtRow(row)
    #     # item.setParent(group_item)
    #
    #     self.myModel.endMoveRows()

    def move_item(self, item, destination_group, row_to_place=None):

        destination_parent = self.index_from_node(destination_group)
        source_parent = self.index_from_node(item.parent)

        if item.parent == destination_group:
            destination_parent = source_parent

        row = item.parent.rowOfChild(item)

        self.beginMoveRows(source_parent, row, row, destination_parent,
                           0 if row_to_place is None else row_to_place)

        if item.parent != destination_group:
            item.parent.removeChildAtRow(row)
            item.setParent(destination_group)
        else:
            destination_group.move_child(row, row_to_place)

        self.endMoveRows()

    def add_items(self, items, parent_item, row=None):
        for item in reversed(items):
            item.setParent(parent_item, row)
        self.insertRows(row if row is not None else len(parent_item.children), len(items),
                        self.index_from_node(parent_item))

    def add_item(self, item, parent_item, row=None):
        """Add child to group a correspondingly updates the view, if row=None, child will be appended at the end."""
        item.setParent(parent_item, row)
        self.insertRow(row if row is not None else len(parent_item.children), self.index_from_node(parent_item))

    def insertRow(self, row, parentIndex):
        return self.insertRows(row, 1, parentIndex)

    def insertRows(self, row, count, parentIndex):
        self.beginInsertRows(parentIndex, row, (row + count - 1))
        self.endInsertRows()
        return True

    def removeRow(self, row, parentIndex):
        return self.removeRows(row, 1, parentIndex)

    def removeRows(self, row, count, parentIndex, update_tristate=True):
        self.beginRemoveRows(parentIndex, row, (row + (count - 1)))
        node = self.node_from_index(parentIndex)
        # print(count)
        del node.children[row:row + count]

        # for i in range(count):
        #     node.removeChildAtRow(row)
        self.endRemoveRows()

        if update_tristate:
            self.update_tristate()
            self.treeView.setup_info()

        return True

    def index(self, row, column, parent):
        node = self.node_from_index(parent)
        return self.createIndex(row, column, node.childAtRow(row))

    def index_from_node(self, node):
        if node is None:
            return QModelIndex()

        parent = node.parent
        return QModelIndex() if parent is None else self.createIndex(parent.rowOfChild(node), 0, node)
        # return self.createIndex(parent.rowOfChild(node) if parent is not None else 0, 0, node)

    def data(self, index, role=None):

        # print("data")

        if role == Qt.DecorationRole:
            return QVariant()

        if role == Qt.TextAlignmentRole:
            return QVariant(int(Qt.AlignTop | Qt.AlignLeft))

        node = self.node_from_index(index)

        if node is None:
            return QVariant()

        if role == Qt.FontRole:
            font = QFont()
            if isinstance(node, SpectrumItemGroup):
                font.setBold(True)

            return QVariant(font)

        if role == Qt.CheckStateRole and index.column() == 0:
            return QVariant(node.check_state)

        if role != Qt.DisplayRole:
            return QVariant()

        if index.column() == 0:
            return QVariant(node.name)

        elif index.column() == 1:
            return QVariant(node.info)

        else:
            return QVariant()

    def columnCount(self, parent):
        return self.columns

    def rowCount(self, parent):
        node = self.node_from_index(parent)
        if node is None:
            return 0
        return len(node)

    def parent(self, index=None):
        if not index.isValid():
            return QModelIndex()

        node = self.node_from_index(index)

        if node is None:
            return QModelIndex()

        parent = node.parent

        if parent is None:
            return QModelIndex()

        grandparent = parent.parent
        if grandparent is None:
            return QModelIndex()
        # print("parent")
        row = grandparent.rowOfChild(parent)

        # row = 0

        # assert row != - 1
        return self.createIndex(row, 0, parent)

    def update_tristate(self):
        # nodes = []
        for group in self.iterate_items(ItemIterator.Groups):
            self._update_tristate(group)
            # nodes += group.children

        # self.checked_changed_signal.emit(nodes)

    def _update_tristate(self, parent):

        if not isinstance(parent, SpectrumItemGroup):
            return

        checked_items = 0
        for item in parent.children:
            if item.isChecked():
                checked_items += 1
        if checked_items == 0:
            parent.check_state = Qt.Unchecked
        elif checked_items < parent.__len__():
            parent.check_state = Qt.PartiallyChecked
        else:
            parent.check_state = Qt.Checked

        idx = self.createIndex(parent.row(), 0, parent)

        self.dataChanged.emit(idx, idx, [Qt.CheckStateRole])

    def setData(self, index, value, role=None):
        if index.column() == 0:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                node.name = value
                self.dataChanged.emit(index, index, [Qt.EditRole])
                self.item_edited_signal.emit(node.isChecked())
                return True

            # this part of code deals with just setting a proper check state of item
            # unchecked, checked and partially checked...
            if role == Qt.CheckStateRole:
                # print("role == Qt.CheckStateRole:")
                node = self.node_from_index(index)
                node.check_state = value

                if isinstance(node, SpectrumItemGroup):
                    if node.__len__() != 0:
                        for item in node.children:
                            item.check_state = value

                        # Index of last child in group
                        idx1 = self.createIndex(node.__len__() - 1, 0, node.children[-1])

                        self.dataChanged.emit(index, idx1, [Qt.CheckStateRole])
                    else:
                        self.dataChanged.emit(index, index, [Qt.CheckStateRole])

                    self.checked_changed_signal.emit(node.children, value)

                elif isinstance(node, SpectrumItem):  # SpectrumItem instance
                    # print("isinstance(node, SpectrumItem):")

                    if node.is_in_group():
                        self._update_tristate(node.parent)

                    self.checked_changed_signal.emit([node], value)
                return True

                # return super(Model, self).setData(index, value, role)

        return super(Model, self).setData(index, value, role)

    def node_from_index(self, index):
        return index.internalPointer() if index.isValid() else self.root


# this is just to keep previous text in field when editing
# http://www.informit.com/articles/article.aspx?p=1405547&seqNum=4
class ItemDelegate(QStyledItemDelegate):

    def setEditorData(self, lineEdit, index):
        previous_text = str(index.model().data(index, Qt.DisplayRole).value())
        lineEdit.setText(previous_text)


class TreeView(QTreeView):

    selecting = False  # this has to be static, method selectionChanged is called before __init__, wtf ??
    deleting = False

    items_deleted_signal = pyqtSignal(bool)

    def __init__(self, parent=None):
        super(TreeView, self).__init__(parent)

        self.myModel = Model(self)
        self.setModel(self.myModel)

        self.setSelectionMode(self.ExtendedSelection)
        self.setSelectionBehavior(QTreeView.SelectRows)
        self.setItemDelegateForColumn(0, ItemDelegate(self))

        self.setUniformRowHeights(True)

        # this is the most important line of code here, this is a result of many hours of unsuccessful tries
        # this sets MoveAction as a default DropAction while dragging and dropping and also if Ctrl is pressed
        # while dragging and dropping, CopyAction is performed instead
        self.setDefaultDropAction(Qt.MoveAction)
        # also this one
        self.setDragDropMode(QAbstractItemView.DragDrop)

        self.dragEnabled()
        self.acceptDrops()
        self.showDropIndicator()

        # self.myModel.dataChanged.connect(self.change)
        self.myModel.dataChanged.connect(lambda topLeftIndex, bottomRightIndex: self.update(topLeftIndex))

        # self.update(topLeftIndex)
        self.myModel.data_dropped_signal.connect(self.save_state)

        self.expandAll()

    def save_state(self):
        self.setup_info()


    # https://stackoverflow.com/questions/50391050/how-to-remove-row-from-qtreeview-using-qabstractitemmodel
    def delete_selected_items(self):
        self.deleting = True

        # for any selected groups, we have to delete them first
        item_was_checked = False

        # --- First remove groups with all of their children ---

        sel_groups = []

        for item in self.myModel.iterate_selected_items(False, True):
            if isinstance(item, SpectrumItemGroup):
                sel_groups.append(item)

        for group in sel_groups:
            if group.check_state == Qt.Checked or group.check_state == Qt.PartiallyChecked:
                item_was_checked = True

            row = self.myModel.root.rowOfChild(group)
            # root_index = self.myModel.indexFromNode(self.myModel.root)

            # QModelIndex of root item is just plain QModelIndex()
            self.myModel.removeRows(row, 1, QModelIndex(), update_tristate=False)

        # --- Remove all remaining SpectrumItems ---

        # try to speed up thing for removing a lot af items, but still, it is kind of slow for > 1000 items...
        # but this is probably problem of Qt itself, because parent() and index() methods are called too many times
        # https://stackoverflow.com/questions/841096/slow-selection-in-qtreeview-why
        while len(self.selectedIndexes()) > 0:
            temp = []
            last_parent = None
            last_row = -1

            for item in self.myModel.iterate_selected_items(True, True):
                # item is SpectrumItem

                current_parent = item.parent

                if last_parent is None:
                    last_parent = current_parent

                current_row = current_parent.rowOfChild(item)

                if item.check_state == Qt.Checked:
                    item_was_checked = True

                if last_row == -1:
                    last_row = current_row

                if current_parent != last_parent or current_row - last_row > 1:
                    break
                else:
                    last_row = current_row
                    last_parent = current_parent
                    temp.append(item)

            if len(temp) == 0:
                continue

            parent = temp[0].parent
            row = parent.rowOfChild(temp[0])
            parent_index = self.myModel.index_from_node(parent)
            self.myModel.removeRows(row, len(temp), parent_index, update_tristate=False)

        self.deleting = False

        self.myModel.update_tristate()
        self.setup_info()

        # emit deleted event and if at least one of the deleted items were checked -> redraw spectra
        self.items_deleted_signal.emit(item_was_checked)
        self.save_state()

    def add_items_to_group(self, items, group=None, row_to_place=None, edit=True, copy_items=False):
        """edit - if place cursor in a new group item to edit the name"""

        if not np.iterable(items):
            raise ValueError("Argument items must be iterable.")

        # add group to the end of root
        group_item = SpectrumItemGroup('', '', parent=self.myModel.root) if group is None else group
        self.myModel.insertRows(self.myModel.root.__len__(), 1, QModelIndex())

        group_item_index = self.myModel.createIndex(len(self.myModel.root), 0, group_item)

        def _move_chunk(parent, first_row, last_row, group_item_index, row_to_put):
            source_parent = self.myModel.index_from_node(parent)

            self.myModel.beginMoveRows(source_parent, first_row, last_row, group_item_index, row_to_put)

            chunk = parent.children[first_row:last_row + 1]  # take the reference
            del parent.children[first_row:last_row + 1]  # delete them from from list

            for child in chunk:
                child.setParent(group_item)

            self.myModel.endMoveRows()

        def _copy_chunk(parent, first_row, last_row, group_item_index, row_to_put):

            self.myModel.beginInsertRows(group_item_index, row_to_put, last_row - first_row + 1)
            chunk = parent.children[first_row:last_row + 1]  # take the reference

            for child in chunk:
                copied_child = child.__copy__()
                copied_child.setParent(group_item)

            self.myModel.endInsertRows()

        chunk_op = _copy_chunk if copy_items else _move_chunk  # either move or copy mode

        last_parent = None
        first_row = -1
        last_row = -1
        i = 0

        for item in items:
            if row_to_place is None:
                row_to_place = item.parent.row() if item.is_in_group() else item.row()

            current_parent = item.parent

            if last_parent is None:
                last_parent = current_parent

            current_row = current_parent.rowOfChild(item)

            if first_row == -1:
                first_row = current_row

            if last_row == -1:
                last_row = current_row

            if current_parent != last_parent or current_row - last_row > 1:  # move chunk of items

                n_items = (last_row - first_row + 1)

                chunk_op(last_parent, first_row, last_row, group_item_index, i - n_items)

                # current row will be changed because we removed n_items
                first_row = current_row if current_parent != last_parent else current_row - n_items
                last_row = first_row
            else:
                last_row = current_row

            last_parent = current_parent
            i += 1

        # process last chunk
        chunk_op(last_parent, first_row, last_row, group_item_index, i - (last_row - first_row + 1))

        # move the group to correct place
        gi = self.myModel.take_item(group_item)
        self.myModel.add_item(gi, self.myModel.root, row_to_place)
        # simple item move did not work for last item in the list, there is some Qt bug there...
        # self.myModel.move_item(group_item, self.myModel.root, row_to_place)

        index = self.myModel.index_from_node(group_item)

        self.expand(index)
        self.myModel.update_tristate()
        self.setup_info()

        # set edit mode of the group item
        if edit:
            self.edit(index)

        self.save_state()

    def add_selected_items_to_group(self, copy=False):

        if len(self.selectedIndexes()) == 0:
            return

        iterator = self.myModel.iterate_selected_items(skip_groups=True,
                                                       skip_childs_in_selected_groups=False)
        self.add_items_to_group(iterator, copy_items=copy)

    def keyPressEvent(self, e):
        pass
        # if e.key() == 16777223:
        #     # pressed delete
        #     self.delete_selected_items()
        #
        # if e.key() == 65 and e.modifiers() == Qt.ControlModifier:
        #     # Ctrl + A
        #     self.selectAll()
        #
        # if e.key() == 86 and e.modifiers() == Qt.ControlModifier:
        #     # Ctrl + V
        #     self.add_selected_items_to_group()
        #
        # if e.key() == 66 and e.modifiers() == Qt.ControlModifier:
        #     # Ctrl + B
        #     self.create_group()
        #
        # if e.key() == 81 and e.modifiers() == Qt.ControlModifier:
        #     # Ctrl + Q
        #     self.check_selected_items()
        #
        # if e.key() == 87 and e.modifiers() == Qt.ControlModifier:
        #     # Ctrl + W
        #     self.uncheck_selected_items()

    def check_selected_items(self):
        if len(self.selectedIndexes()) == 0:
            return

        items = []

        for item in self.myModel.iterate_selected_items(False, False):
            item.check_state = Qt.Checked
            items.append(item)

        self.myModel.update_tristate()
        self.myModel.dataChanged.emit(QModelIndex(), QModelIndex())
        self.myModel.checked_changed_signal.emit(items, True)

    def uncheck_selected_items(self):
        if len(self.selectedIndexes()) == 0:
            return

        items = []

        for item in self.myModel.iterate_selected_items(False, False):
            item.check_state = Qt.Unchecked
            items.append(item)

        self.myModel.update_tristate()
        self.myModel.dataChanged.emit(QModelIndex(), QModelIndex())
        self.myModel.checked_changed_signal.emit(items, False)

    def uncheck_all(self):

        items = []
        for item in self.myModel.iterate_items(ItemIterator.All):
            item.check_state = Qt.Unchecked
            items.append(item)

        self.myModel.dataChanged.emit(QModelIndex(), QModelIndex())
        self.myModel.all_unchecked_signal.emit()

    def create_group(self):
        group_item = SpectrumItemGroup('', parent=self.myModel.root)
        self.myModel.insertRows(self.myModel.root.__len__(), 1, QModelIndex())
        self.edit(self.myModel.index_from_node(group_item))

    def sort_tree_view(self, sort_groups=False, ascending=True):

        if self.myModel.root.__len__() == 0:
            return
        items = []

        self.myModel.layoutAboutToBeChanged.emit([])
        step = 1 if ascending else -1
        self.myModel.root.children = sorted(self.myModel.root.children, key=lambda child: child.name)[::step]
        self.myModel.layoutChanged.emit([])

        items.append(self.myModel.root.children)

        if sort_groups:
            self.myModel.layoutAboutToBeChanged.emit([])
            for group in self.myModel.iterate_items(ItemIterator.Groups):
                if group.__len__() == 0:
                    continue
                group.children = sorted(group.children, key=lambda child: child.name)[::step]
                items.append(group.children)

            self.myModel.layoutChanged.emit([])

        self.clearSelection()
        self.myModel.items_sorted_signal.emit(items)
        self.save_state()

    def sort_selected_group(self, ascending=True):
        self.sort_group(self.myModel.node_from_index(self.currentIndex()), ascending)

    def sort_group(self, group_item, ascending=True):
        # help from documentation for emitting signals:
        # http://doc.qt.io/qt-5/qabstractitemmodel.html#layoutAboutToBeChanged
        if group_item.__len__() == 0:
            return

        items = []

        self.myModel.layoutAboutToBeChanged.emit([])
        step = 1 if ascending else -1
        group_item.children = sorted(group_item.children, key=lambda child: child.name)[::step]
        items.append(group_item.children)
        self.myModel.layoutChanged.emit([])

        self.clearSelection()
        self.myModel.items_sorted_signal.emit(items)

    def ungroup_selected_group(self):
        self.ungroup(self.myModel.node_from_index(self.currentIndex()))

    def ungroup(self, group_item):
        assert isinstance(group_item, SpectrumItemGroup)

        if group_item.__len__() == 0:
            return

        children = self.myModel.take_all_children(group_item)
        row = group_item.row()
        self.myModel.add_items(children, self.myModel.root, row + 1)

        del group_item

        self.myModel.removeRow(row, QModelIndex())

        self.myModel.items_ungrouped_signal.emit(children)
        self.save_state()

    # def change(self, topLeftIndex, bottomRightIndex):
    #     self.update(topLeftIndex)

    def clear(self):
        """Removes all items from the TreeView"""
        self.myModel.removeRows(0, self.top_level_items_count(), QModelIndex())

    def setup_info(self):

        for i, item in enumerate(self.myModel.root.children):

            # for all items
            if isinstance(item, SpectrumItemGroup):
                it_len = item.__len__()
                item.info = "[{}] | {} item{}".format(i, it_len, '' if it_len == 1 else 's')

                for j, child in enumerate(item.children):
                    child.info = "[{}][{}] | {}; {} ({:.4g}, {:.4g})".format(i, j,
                                                                             child.length(),
                                                                             child.spacing(),
                                                                             child.x.min(),
                                                                             child.x.max())

            else:
                item.info = "[{}] | {}; {} ({:.4g}, {:.4g})".format(i,
                                                                    item.length(),
                                                                    item.spacing(),
                                                                    item.x.min(),
                                                                    item.x.max())


    def update_view(self):
        self.myModel.dataChanged.emit(QModelIndex(), QModelIndex())
        self.save_state()

    # def select_item(self, item):
    #     self.selecting = True
    #
    #     flags = QItemSelectionModel.Select
    #     selection = QItemSelection()
    #
    #     start_index = self.myModel.createIndex(item.row(), 0, item)
    #     end_index = self.myModel.createIndex(item.row(), 1, item)
    #
    #     selection.select(start_index, end_index)
    #     self.selectionModel().select(selection, flags)
    #
    #     self.selecting = False

    def selectionChanged(self, selected, deselected):

        super(TreeView, self).selectionChanged(selected, deselected)

        if self.selecting or self.deleting:
            return

        self.selecting = True

        # select all children if group was selected
        for item in self.myModel.iterate_selected_items(skip_groups=False, skip_childs_in_selected_groups=True):
            # help from https://stackoverflow.com/questions/47271494/set-selection-from-a-list-of-indexes-in-a-qtreeview
            if isinstance(item, SpectrumItemGroup):
                # mod = self.model()
                # columns = mod.columnCount() - 1

                length = item.__len__()

                if length == 0:
                    continue

                flags = QItemSelectionModel.Select
                selection = QItemSelection()

                start_index = self.myModel.createIndex(0, 0, item.children[0])
                end_index = self.myModel.createIndex(length - 1, 1, item.children[-1])

                selection.select(start_index, end_index)
                self.selectionModel().select(selection, flags)

        self.selecting = False

    def dragEnterEvent(self, e):
        if e.source():
            super(TreeView, self).dragEnterEvent(e)
            # print("dragEnterEvent - source")
        else:
            m = e.mimeData()
            if m.hasUrls():
                for url in m.urls():
                    if url.isLocalFile():
                        e.accept()
                return
            elif m.hasText():
                e.accept()
                return
            elif m.hasFormat("XML Spreadsheet"):
                e.accept()
                return
            elif m.hasFormat(PyMimeData.MIME_TYPE):
                super(TreeView, self).dragEnterEvent(e)
                # print("dragEnterEvent - from another app")
                # e.accept()
                return

            e.ignore()

    def dropEvent(self, e):

        if e.source():
            super(TreeView, self).dropEvent(e)
        else:
            m = e.mimeData()

            if m.hasFormat(PyMimeData.MIME_TYPE):
                super(TreeView, self).dropEvent(e)
                return

            if m.hasUrls():
                filepaths = []
                for url in m.urls():
                    # print(url.toString(), url.path())
                    if url.isLocalFile():
                        # e.accept()
                        if platform == "win32":
                            path = url.path()[1:]  # and  .replace('/', '\\')
                        else:
                            path = url.path()
                        filepaths.append(path)
                self.import_files(filepaths)
                return

            if m.hasFormat("XML Spreadsheet"):
                self.parse_XML_Spreadsheet(m.data("XML Spreadsheet").data())
                return
            e.ignore()

    @abstractmethod
    def import_files(self, filepaths):
        raise NotImplementedError

    @abstractmethod
    def parse_XML_Spreadsheet(self, byte_data):
        raise NotImplementedError

    def expanded(self, index):
        for column in range(self.model().columnCount(QModelIndex())):
            self.resizeColumnToContents(column)

    def top_level_items_count(self):
        return self.myModel.root.__len__()


if __name__ == "__main__":
    def my_exception_hook(exctype, value, traceback):
        # Print the error and traceback
        print(exctype, value, traceback)
        # Call the normal Exception hook after
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)


    from PyQt5.QtWidgets import QApplication

    sys._excepthook = sys.excepthook

    # Set the exception hook to our wrapping function
    sys.excepthook = my_exception_hook

    app = QApplication(sys.argv)
    MainWindow = TreeView()
    # ui = Ui_MainWindow()
    # ui.setupUi(MainWindow)
    MainWindow.show()
    # sys.exit(app.exec_())

    cProfile.run('app.exec_()', 'profdata')
    p = pstats.Stats('profdata')
    p.sort_stats('time').print_stats()



from ..qt_task import Task
from ..logger import Logger
import time
from lmfit import Minimizer
# from multiprocessing import Process
import numpy as np
from ..dialogs.fitresult import FitResult


class TaskFit(Task):
    def __init__(self, fw, model, method='least_squares', model_options=None, parent=None):
        super(TaskFit, self).__init__(parent)
        self.fw = fw  # fit widget
        self.model = model
        self.method = method
        self.model_options = {} if model_options is None else model_options

        self.result = None
        self.minimizer = None

    def preRun(self):
        self.fw.btnFit.setText('Cancel')

    def run(self):
        Logger.debug('Fitting...')
        self.run_fit()

    def run_fit(self):

        start_time = time.perf_counter()

        self.minimizer = Minimizer(self.model.residuals, self.model.params,
                                   iter_cb=lambda params, it, resid, *args, **kws: self.isInterruptionRequested())

        self.result = self.minimizer.minimize(method=self.method, **self.model_options)  # fit

        end_time = time.perf_counter()
        Logger.debug(end_time - start_time, 's for fitting')

    def postRun(self):  # after run has finished
        values_errors = np.zeros((len(self.result.params), 2), dtype=np.float64)
        for i, (p, new_p) in enumerate(zip(self.model.params.values(), self.result.params.values())):
            p.value = new_p.value  # update fitted parameters
            values_errors[i, 0] = p.value
            values_errors[i, 1] = p.stderr if p.stderr is not None else 0

        x_vals, fits, residuals = self.model.simulate()

        self.fw.setup_fields()
        self.fw.plot_fits(x_vals, fits, residuals)

        self.fw.fit_result = FitResult(self.result, self.minimizer, values_errors, self.model)
        self.fw.btnFit.setText('Fit')





# This is modified QLineEdit with added lost_focus signal

from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtCore import pyqtSignal


class MyLineEdit(QLineEdit):
    focus_lost = pyqtSignal()

    def focusOutEvent(self, e):
        self.focus_lost.emit()
        super(MyLineEdit, self).focusOutEvent(e)



import numpy as np
from typing import Iterable
import os

import matplotlib as mpl
import matplotlib.pyplot as plt  # we plot graphs with this library
from matplotlib import cm
from matplotlib.ticker import *
import matplotlib.gridspec as gridspec

from matplotlib import colors as c

# from copy import deepcopy
from PyQt5.QtWidgets import QApplication
from scipy.linalg import lstsq

from spectramanipulator.settings import Settings
from spectramanipulator.spectrum import fi, Spectrum, SpectrumList, group2mat
from scipy.integrate import simps, cumtrapz
from scipy.stats import linregress
from uncertainties import ufloat, unumpy


import functools

# for backward compatibility of smpj files
ItemList = list

WL_LABEL = 'Wavelength / nm'
WN_LABEL = "Wavenumber / $10^{4}$ cm$^{-1}$"


# needed for correctly display tics for symlog scale
class MinorSymLogLocator(Locator):
    """
    Dynamically find minor tick positions based on the positions of
    major ticks for a symlog scaling.
    """

    def __init__(self, linthresh, nints=10):
        """
        Ticks will be placed between the major ticks.
        The placement is linear for x between -linthresh and linthresh,
        otherwise its logarithmically. nints gives the number of
        intervals that will be bounded by the minor ticks.
        """
        self.linthresh = linthresh
        self.nintervals = nints

    def __call__(self):
        # Return the locations of the ticks
        majorlocs = self.axis.get_majorticklocs()

        if len(majorlocs) == 1:
            return self.raise_if_exceeds(np.array([]))

        # add temporary major tick locs at either end of the current range
        # to fill in minor tick gaps
        dmlower = majorlocs[1] - majorlocs[0]  # major tick difference at lower end
        dmupper = majorlocs[-1] - majorlocs[-2]  # major tick difference at upper e	nd

        # add temporary major tick location at the lower end
        if majorlocs[0] != 0. and ((majorlocs[0] != self.linthresh and dmlower > self.linthresh) or (
                dmlower == self.linthresh and majorlocs[0] < 0)):
            majorlocs = np.insert(majorlocs, 0, majorlocs[0] * 10.)
        else:
            majorlocs = np.insert(majorlocs, 0, majorlocs[0] - self.linthresh)

        # add temporary major tick location at the upper end
        if majorlocs[-1] != 0. and ((np.abs(majorlocs[-1]) != self.linthresh and dmupper > self.linthresh) or (
                dmupper == self.linthresh and majorlocs[-1] > 0)):
            majorlocs = np.append(majorlocs, majorlocs[-1] * 10.)
        else:
            majorlocs = np.append(majorlocs, majorlocs[-1] + self.linthresh)

        # iterate through minor locs
        minorlocs = []

        # handle the lowest part
        for i in range(1, len(majorlocs)):
            majorstep = majorlocs[i] - majorlocs[i - 1]
            if abs(majorlocs[i - 1] + majorstep / 2) < self.linthresh:
                ndivs = self.nintervals
            else:
                ndivs = self.nintervals - 1.

            minorstep = majorstep / ndivs
            locs = np.arange(majorlocs[i - 1], majorlocs[i], minorstep)[1:]
            minorlocs.extend(locs)

        return self.raise_if_exceeds(np.array(minorlocs))

    def tick_values(self, vmin, vmax):
        raise NotImplementedError('Cannot get tick locations for a '
                                  '%s type.' % type(self))


def setup_wavenumber_axis(ax, x_label=WN_LABEL,
                          x_major_locator=None, x_minor_locator=AutoMinorLocator(5), factor=1e3):
    secondary_ax = ax.secondary_xaxis('top', functions=(lambda x: factor / x, lambda x: 1 / (factor * x)))

    secondary_ax.tick_params(which='major', direction='in')
    secondary_ax.tick_params(which='minor', direction='in')

    if x_major_locator:
        secondary_ax.xaxis.set_major_locator(x_major_locator)

    if x_minor_locator:
        secondary_ax.xaxis.set_minor_locator(x_minor_locator)

    secondary_ax.set_xlabel(x_label)

    return secondary_ax


def set_main_axis(ax, x_label=WL_LABEL, y_label="Absorbance", xlim=(None, None), ylim=(None, None),
                  x_major_locator=None, x_minor_locator=None, y_major_locator=None, y_minor_locator=None):
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if xlim[0] is not None:
        ax.set_xlim(xlim)
    if ylim[0] is not None:
        ax.set_ylim(ylim)

    if x_major_locator:
        ax.xaxis.set_major_locator(x_major_locator)

    if x_minor_locator:
        ax.xaxis.set_minor_locator(x_minor_locator)

    if y_major_locator:
        ax.yaxis.set_major_locator(y_major_locator)

    if y_minor_locator:
        ax.yaxis.set_minor_locator(y_minor_locator)

    ax.tick_params(axis='both', which='major', direction='in')
    ax.tick_params(axis='both', which='minor', direction='in')


def _transform_func(transform=lambda y: y):
    def decorator(fn):
        @functools.wraps(fn)
        def func(item):
            fn_name = fn.__name__
            if isinstance(item, Spectrum):
                y_data = transform(item.data[:, 1])
                return Spectrum.from_xy_values(item.data[:, 0], np.nan_to_num(y_data), name=f'{fn_name}({item.name})')
            elif isinstance(item, SpectrumList):
                sl = SpectrumList(name=f'{fn_name}({item.name})')
                for sp in item:
                    y_data = transform(sp.data[:, 1])
                    new_sp = Spectrum.from_xy_values(sp.data[:, 0], np.nan_to_num(y_data), name=f'{fn_name}({sp.name})')
                    sl.children.append(new_sp)
                return sl
            else:
                return transform(item)  # this may not be implemented, but for numbers and ndarrays it will work
        return func
    return decorator


@_transform_func(lambda y: np.exp(y))
def exp(item):
    """
    Calculates the exponential of y values and returns a new Spectrum/SpectrumList.
    """
    pass


@_transform_func(lambda y: np.log10(y))
def log10(item):
    """
    Calculates the decadic logarithm of y values and returns a new Spectrum/SpectrumList.
    """
    pass


@_transform_func(lambda y: np.log(y))
def log(item):
    """
    Calculates the natural logarithm of y values and returns a new Spectrum/SpectrumList
    """
    pass


@_transform_func(lambda y: -np.log10(-y))
def T2A_LFP(item):
    """
    Performs the transmittance to absorbance conversion for nano kinetics,
    y_transformed = -log10(-y)."""
    pass

#
# def add_to_list(spectra):
#     """
#     Copies all spectra and imports them to the Tree Widget.
#
#     Parameters
#     ----------
#     spectra : {:class:`Spectrum`, :class:`SpectrumList`, list, list of lists, SpectrumItemGroup, SpectrumItem}
#         The input spectra to be added into Tree Widget.
#     """
#
#     if UserNamespace.instance is not None:
#         UserNamespace.instance.tw.add_to_list(spectra)


# def load_kinetics(spectra):
#     """
#     Copies all spectra and imports them to the Tree Widget.
#
#     Parameters
#     ----------
#     spectra : {:class:`Spectrum`, :class:`SpectrumList`, list, list of lists, SpectrumItemGroup, SpectrumItem}
#         The input spectra to be added into Tree Widget.
#     """
#
#     if UserNamespace.instance is not None:
#         UserNamespace.instance.tw.add_to_list(spectra)



def import_files(filepaths):
    """
    Imports the filepaths and add to Tree Widget

    Parameters
    ----------
    filepaths : list of strs or str
        List of filepaths to import.
    """
    if UserNamespace.instance is not None:
        UserNamespace.instance.main.tree_widget.import_files(filepaths)


def set_xy_range(x0=None, x1=None, y0=None, y1=None, padding=0):
    """
    Changes the x and y ranges of scene in Plot Widget.

    Parameters
    ----------
    x0 : {int, float, None}
        New fist x value. If None, old value is kept.
    x1 : {int, float, None}
        New last x value. If None, old value is kept.
    y0 : {int, float, None}
        New fist y value. If None, old value is kept.
    y1 : {int, float, None}
        New last y value. If None, old value is kept.
    padding : {int, float}
        Sets the padding around the choosed rectangle. If 0, no padding will be used.
    """
    plot_widget = UserNamespace.instance.main.grpView

    x_range, y_range = plot_widget.plotItem.getViewBox().viewRange()

    plot_widget.plotItem.getViewBox().setXRange(x_range[0] if x0 is None else x0,
                                                x_range[1] if x1 is None else x1,
                                                padding=padding)

    plot_widget.plotItem.getViewBox().setYRange(y_range[0] if y0 is None else y0,
                                                y_range[1] if y1 is None else y1,
                                                padding=padding)


def set_default_HSV_color_scheme():
    """Sets the default values for HSV color scheme."""
    Settings.hues = 9
    Settings.values = 1
    Settings.maxValue = 255
    Settings.minValue = 150
    Settings.maxHue = 360
    Settings.minHue = 0
    Settings.sat = 255
    Settings.alpha = 255

    if Settings.HSV_color_scheme:
        redraw_all_spectra()


def set_HSV_color_scheme(active=True, **kwargs):
    """Set the options for HSV color scheme and whether the scheme is active.

    Options
    -------
    ================  =================================================================================
    *active* (bool)   True for setting the scheme active, False for not (default color scheme will be
                      used).
    *hues* (int)      The number of hues that will be repeating, default 9.
    *values* (int)    The number of values/brightnesses that will be repeating, default 1.
    *minValue* (int)  A minimum value/brightness, this can be <0, 255>, default 150.
    *maxValue* (int)  A maximum value/brightness, this can be <0, 255>, default 255.
    *minHue* (int)    A minimum hue, this can be <0, 360>, default 0
    *maxHue* (int)    A maximum hue, this can be <0, 360>, default 360
    *sat* (int)       The saturation value, this can be <0, 255>, default 255
    *alpha* (int)     The transparency value, this can be <0, 255>, default 255
    ================  =================================================================================
    """
    hues = kwargs.get('hues', None)
    values = kwargs.get('values', None)
    maxValue = kwargs.get('maxValue', None)
    minValue = kwargs.get('minValue', None)
    maxHue = kwargs.get('maxHue', None)
    minHue = kwargs.get('minHue', None)
    sat = kwargs.get('sat', None)
    alpha = kwargs.get('alpha', None)

    Settings.HSV_color_scheme = active
    Settings.hues = hues if hues is not None else Settings.hues
    Settings.values = values if values is not None else Settings.values
    Settings.maxValue = maxValue if maxValue is not None else Settings.maxValue
    Settings.minValue = minValue if minValue is not None else Settings.minValue
    Settings.maxHue = maxHue if maxHue is not None else Settings.maxHue
    Settings.minHue = minHue if minHue is not None else Settings.minHue
    Settings.sat = sat if sat is not None else Settings.sat
    Settings.alpha = alpha if alpha is not None else Settings.alpha

    redraw_all_spectra()


def copy_to_clipboard(array, delimiter='\t', decimal_sep='.', new_line='\n'):
    """Copies the *array* of numbers into clipboard. This can be then pasted to Excel for example.

    Parameters
    ----------
    array : {array_like, iterable}
        Array of values. Can be 1D or 2D array
    delimiter : str
        Delimiter between numbers, default tabulator '\\\\t'
    decimal_sep : str
        Decimal separator, default '.'
    new_line : str
        New line character, default '\\\\n'
    """
    if not isinstance(array, (np.ndarray, Iterable, list, tuple)):
        raise ValueError(f"Cannot copy {type(array)} to clipboard.")

    try:
        text = new_line.join(delimiter.join(str(num).replace('.', decimal_sep) for num in row) for row in array)
    except:  # the second dimension is not iterable, we probably got only 1D array, so lets put into clipboard only this
        text = delimiter.join(str(num).replace('.', decimal_sep) for num in array)

    cb = QApplication.clipboard()
    cb.clear(mode=cb.Clipboard)
    cb.setText(text, mode=cb.Clipboard)


def update_view():
    """Updates the Tree Widget."""
    if UserNamespace.instance is None:
        return

    mw = UserNamespace.instance.main

    mw.tree_widget.update_view()
    mw.tree_widget.setup_info()


def redraw_all_spectra():
    """Redraws all spectra."""
    if UserNamespace.instance is None:
        return

    mw = UserNamespace.instance.main

    mw.redraw_all_spectra()

###  Calculation of epsilon from concentration-dependent absorption spectra, the name of the spectra must contain
###  real concentration, the spectra must me ordered from lowest to highest concentration

def _get_C(group):
    """Returns parsed names to floats from a group"""
    x_vals_temp = []
    for sp in group:
        try:
            x_vals_temp.append(float(sp.name.replace(',', '.').strip()))
        except ValueError:
            raise ValueError("Names of spectra cannot be parsed to float.")
    return np.asarray(x_vals_temp, dtype=np.float64)

def _get_D(group):
    D = group[0].data[:, 1]
    for i in range(1, len(group)):
        D = np.vstack((D, group[i].data[:, 1]))
    return D

#
# def calc_Eps(group):
#     C = _get_C(group)
#     D = _get_D(group)
#     wls = group[0].data[:, 0]
#
#     # C needs to be changed to column vector
#     ST = lstsq(C.reshape(-1, 1), D)[0]
#
#     # add a spectrum to list
#     Spectrum.from_xy_values(wls, ST.flatten(), name=group.name + '-epsilon').add_to_list()


def rename_times(group, decimal_places=1):
    """Renames the group that has names in seconds. Changes for minutes for 60s <= time < 1 hour to minutes and
    time >= 1 hour to hours."""

    parsed_times = []
    times = _get_C(group)

    for time in times:
        unit = ' s'
        if time >= 3600:
            time /= 3600
            unit = ' h'
        elif 60 <= time < 3600:
            time /= 60
            unit = ' min'

        time = np.round(time, decimal_places)
        parsed_times.append(f'{time}{unit}')

    group.set_names(parsed_times)


# def load_kinetics(dir_name, spectra_dir_name='spectra', times_fname='times.txt', blank_spectrum='blank.dx', dt=None,
#                   b_corr=None, cut=None, corr_to_zero_time=True):
#     """Given a directory name that contains folders of individual experiments, it loads all kinetics.
#        each experiment folder must contain folder spectra (or defined in spectra_dir_name arg.)
#         if blank is given, it will be subtracted from all spectra, times.txt will contain
#         times for all spectra, optional baseline correction and cut can be done.
#
#     Folder structure:
#         [dir_name]
#             [exp1_dir]
#                 [spectra]
#                     01.dx (or .csv or .txt)
#                     02.dx
#                     ...
#                 times.txt (optional)
#                 blank.dx (optional)
#             [exp2_dir]
#                 ...
#             ...
#     """
#
#     if UserNamespace.instance is None:
#         return
#
#     if not os.path.isdir(dir_name):
#         raise ValueError(f'{dir_name}  does not exist!')
#
#     for item in os.listdir(dir_name):
#         path = os.path.join(dir_name, item)
#         if not os.path.isdir(path):
#             continue
#
#         load_kinetic(path, spectra_dir_name=spectra_dir_name, times_fname=times_fname, blank_spectrum=blank_spectrum,
#                      dt=dt, b_corr=b_corr, cut=cut, corr_to_zero_time=corr_to_zero_time)
#
#
# def load_kinetic(dir_name, spectra_dir_name='spectra', times_fname='times.txt', blank_spectrum='blank.dx', dt=None,
#                   b_corr=None, cut=None, corr_to_zero_time=True):
#     """Given a directory name, it loads all spectra in dir named "spectra" - func. arg.,
#     if blank is given, it will be subtracted from all spectra, times.txt will contain
#     times for all spectra, optional baseline correction and cut can be done.
#
#     Folder structure:
#         [dir_name]
#             [spectra]
#                 01.dx
#                 02.dx
#                 ...
#             times.txt (optional)
#             blank.dx (optional)
#     """
#
#     if UserNamespace.instance is None:
#         return
#
#     tw = UserNamespace.instance.main.tree_widget
#     root = tw.myModel.root  # item in IPython console
#
#     if not os.path.isdir(dir_name):
#         raise ValueError(f'{dir_name}  does not exist!')
#
#     spectra_path = os.path.join(dir_name, spectra_dir_name)
#
#     if not os.path.isdir(spectra_path):
#         raise ValueError(f'{spectra_dir_name}  does not exist in {dir_name}!')
#
#     spectras = [os.path.join(spectra_path, filename) for filename in os.listdir(spectra_path)]
#
#     n_items_before = root.__len__()
#     tw.import_files(spectras)
#     n_spectra = root.__len__() - n_items_before
#
#     tw.add_items_to_group(root[n_items_before:], edit=False)  # add loaded spectra to group
#     root[n_items_before].name = f'raw [{os.path.split(dir_name)[1]}]'  # set name of a group
#
#     times = np.asarray([dt * i for i in range(n_spectra)]) if dt is not None else None
#     # idx_add = 0
#     group_idx = n_items_before
#     blank_used = False
#
#     # load explicit times
#     times_fpath = os.path.join(dir_name, times_fname)
#     if os.path.isfile(times_fpath):
#         tw.import_files(times_fpath)
#         # idx_add += 1
#         if times is None:
#             times = root[-1].data[:, 0].copy()
#             if corr_to_zero_time:
#                 times -= times[0]
#
#         # push times variable to the console
#         UserNamespace.instance.main.console.push_variables(
#             {
#                 'times': times
#             }
#         )
#
#     if times is not None:
#         root[group_idx].set_names(times)
#
#     # load blank spectrum if available
#     blank_fpath = os.path.join(dir_name, blank_spectrum)
#     if os.path.isfile(blank_fpath):
#         last_idx = root.__len__() - 1
#         tw.import_files(blank_fpath)
#         add_to_list(root[group_idx] - root[last_idx + 1])
#         if times is not None:
#             root[-1].set_names(times)
#         blank_used = True
#
#     corr_idx = -1 if blank_used else group_idx
#
#     if b_corr is not None:
#         root[corr_idx].baseline_correct(*b_corr)
#         root[corr_idx].name += 'bcorr'
#     if cut is not None:
#         root[corr_idx].cut(*cut)
#         root[corr_idx].name += 'cut'
#
#     # return times
#

def _setup_wavenumber_axis(ax, x_label=WN_LABEL,
                          x_major_locator=None, x_minor_locator=AutoMinorLocator(5), factor=1e3):
    secondary_ax = ax.secondary_xaxis('top', functions=(lambda x: factor / x, lambda x: 1 / (factor * x)))

    secondary_ax.tick_params(which='major', direction='in')
    secondary_ax.tick_params(which='minor', direction='in')

    if x_major_locator:
        secondary_ax.xaxis.set_major_locator(x_major_locator)

    if x_minor_locator:
        secondary_ax.xaxis.set_minor_locator(x_minor_locator)

    secondary_ax.set_xlabel(x_label)

    return secondary_ax


def _set_main_axis(ax, x_label=WL_LABEL, y_label="Absorbance", xlim=(None, None), ylim=(None, None),
                  x_major_locator=None, x_minor_locator=None, y_major_locator=None, y_minor_locator=None,
                   direction='in'):
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if xlim[0] is not None:
        ax.set_xlim(xlim)
    if ylim[0] is not None:
        ax.set_ylim(ylim)

    if x_major_locator:
        ax.xaxis.set_major_locator(x_major_locator)

    if x_minor_locator:
        ax.xaxis.set_minor_locator(x_minor_locator)

    if y_major_locator:
        ax.yaxis.set_major_locator(y_major_locator)

    if y_minor_locator:
        ax.yaxis.set_minor_locator(y_minor_locator)

    ax.tick_params(axis='both', which='major', direction=direction)
    ax.tick_params(axis='both', which='minor', direction=direction)


def plot_kinetics(group_item, n_spectra=50, linscale=1, linthresh=100, cmap='jet_r',
                  major_ticks_labels=(100, 1000), emph_t=(0, 200, 1000), inset_loc=(0.75, 0.1, 0.03, 0.8),
                  colorbar_label='Time / s', lw=0.5, alpha=0.5, fig_size=(5, 4), y_label='Absorbance', x_label=WL_LABEL,
                  x_lim=(230, 600), filepath=None, dpi=500, transparent=True):

    t = np.asarray(group_item.get_names(), dtype=np.float64)
    w = group_item[0].data[:, 0]

    fig, ax1 = plt.subplots(1, 1, figsize=fig_size)

    _set_main_axis(ax1, x_label=x_label, y_label=y_label, xlim=x_lim, x_minor_locator=None, y_minor_locator=None)
    _ = _setup_wavenumber_axis(ax1)

    cmap = cm.get_cmap(cmap)
    norm = mpl.colors.SymLogNorm(vmin=t[0], vmax=t[-1], linscale=linscale, linthresh=linthresh, base=10, clip=True)

    tsb_idxs = fi(t, emph_t)
    ts_real = np.round(t[tsb_idxs])

    x_space = np.linspace(0, 1, n_spectra, endpoint=True, dtype=np.float64)

    t_idx_space = fi(t, norm.inverse(x_space))
    t_idx_space = np.sort(np.asarray(list(set(t_idx_space).union(set(tsb_idxs)))))

    for i in t_idx_space:
        x_real = norm(t[i])
        x_real = 0 if np.ma.is_masked(x_real) else x_real
        ax1.plot(w, group_item[i].data[:, 1], color=cmap(x_real),
                 lw=1.5 if i in tsb_idxs else lw,
                 alpha=1 if i in tsb_idxs else alpha,
                 zorder=1 if i in tsb_idxs else 0)

    cbaxes = ax1.inset_axes(inset_loc)

    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cbar = plt.colorbar(sm, cax=cbaxes, orientation='vertical',
                        format=mpl.ticker.ScalarFormatter(),
                        label=colorbar_label)

    cbaxes.invert_yaxis()

    minor_ticks = [10, 20, 30, 40, 50, 60, 70, 80, 90, 200, 300, 400, 500, 600, 700, 800, 900] + list(
        np.arange(2e3, t[-1], 1e3))
    cbaxes.yaxis.set_ticks(cbar._locate(minor_ticks), minor=True)

    major_ticks = np.sort(np.hstack((np.asarray([100, 1000]), ts_real)))
    major_ticks_labels = np.sort(np.hstack((np.asarray(major_ticks_labels), ts_real)))

    cbaxes.yaxis.set_ticks(cbar._locate(major_ticks), minor=False)
    cbaxes.set_yticklabels([(f'{num:0.0f}' if num in major_ticks_labels else "") for num in major_ticks])

    for ytick, ytick_label, _t in zip(cbaxes.yaxis.get_major_ticks(), cbaxes.get_yticklabels(), major_ticks):
        if _t in ts_real:
            color = cmap(norm(_t))
            ytick_label.set_color(color)
            ytick_label.set_fontweight('bold')
            ytick.tick2line.set_color(color)
            ytick.tick2line.set_markersize(5)
            # ytick.tick2line.set_markeredgewidth(2)

    if filepath:
        ext = os.path.splitext(filepath)[1].lower()[1:]
        plt.savefig(fname=filepath, format=ext, transparent=transparent, dpi=dpi)

    plt.show()


def plot_kinetics_no_colorbar(group_item, x_lim=(None, None), y_lim=(None, None), slice_to_plot=slice(0, -1, 5),
                              x_label='Time / s', y_label='$A$', cmap='jet', darkens_factor_cmap=1, colors=None,
                              x_major_locator=None, x_minor_locator=None,
                              y_major_locator=None, y_minor_locator=None,
                              add_wn_axis=True, lw=1.5, ls='-', plot_zero_line=True,
                              label_format_fcn=lambda name: name,
                              legend_loc='best', legend_spacing=0.2, legend_columns=1, legend_column_spacing=2,
                              legend_entry_prefix='pH = ', legend_entry_postfix='', plot_legend_line=True,
                              fig_size=(5.5, 4.5),
                              dpi=500, filepath=None, transparent=True):

    fig, ax = plt.subplots(1, 1, figsize=fig_size)

    x = group_item[0].data[:, 0]
    sel_items = group_item[slice_to_plot]

    x_range = (x_lim[0] if x_lim[0] is not None else x[0], x_lim[1] if x_lim[1] is not None else x[-1])

    set_main_axis(ax, x_label=x_label, y_label=y_label, xlim=x_range, ylim=y_lim,
                  x_major_locator=x_major_locator, x_minor_locator=x_minor_locator,
                  y_major_locator=y_major_locator, y_minor_locator=y_minor_locator)

    if add_wn_axis:
        _ = setup_wavenumber_axis(ax, x_major_locator=MultipleLocator(0.5))

    _cmap = cm.get_cmap(cmap, len(sel_items))

    if plot_zero_line:

        ax.axhline(0, x_range[0], x_range[1], ls='--', color='black', lw=1)

    for i, item in enumerate(sel_items):
        if colors is None:
            color = np.asarray(c.to_rgb(_cmap(i))) * darkens_factor_cmap
            color[color > 1] = 1
        else:
            color = colors[i % len(colors)]

        ax.plot(item.x, item.y, color=color, lw=lw, ls=ls,
                label=f'{legend_entry_prefix}{label_format_fcn(item.name)}{legend_entry_postfix}')

    l = ax.legend(loc=legend_loc, frameon=False, labelspacing=legend_spacing, ncol=legend_columns,
                  handlelength=None if plot_legend_line else 0, handletextpad=None if plot_legend_line else 0,
                  columnspacing=legend_column_spacing)

    for i, text in enumerate(l.get_texts()):
        # text.set_ha('right')
        text.set_color(_cmap(i))

    ax.set_axisbelow(False)
    ax.yaxis.set_ticks_position('both')

    plt.tight_layout()

    if filepath:
        ext = os.path.splitext(filepath)[1].lower()[1:]
        plt.savefig(fname=filepath, format=ext, transparent=transparent, dpi=dpi)
    else:
        plt.show()


def plot_EEMs(EEM_group_items: list, n_rows: int = None, n_cols: int = None, log_z: bool = False, transform2wavenumber=True,
              fig_size_one_graph=(5.5, 4), x_lim=(None, None), y_lim=(None, None), z_lim=(1, None),  filepath=None, dpi=500,
              transparent=False, show_title=True, cmap='hot_r', z_label='Counts', x_major_locators=(None, None), x_minor_locators=(None, None),
              y_major_locators=(None, None), y_minor_locators=(None, None)):

    """This will assume that excitation wavelengths are used as names for individual spectra
    x a y lims are in given wavelengths, despite the possible recalculation to wavenumber."""

    n = len(EEM_group_items)  # number of EEMs to plot

    if n_rows is None and n_cols is None:  # estimate the n_rows and n_cols from the sqrt of number of graphs
        sqrt = n ** 0.5
        n_rows = int(np.ceil(sqrt))
        n_cols = int(sqrt)
    elif n_rows is None and n_cols is not None:
        n_rows = int(np.ceil(n / n_cols))
    elif n_rows is not None and n_cols is None:
        n_cols = int(np.ceil(n / n_rows))

    # assert n_rows * n_cols >= n  # not necessary, if the condition is not valid, fewer plots will be plotted

    fig, axes = plt.subplots(n_rows, n_cols, figsize=(fig_size_one_graph[0] * n_cols, fig_size_one_graph[1] * n_rows))

    axes = axes.flatten() if np.iterable(axes) else [axes]

    t2w = lambda x: 1e3 / x  # function that transforms wavelength into 10^4 cm-1

    for ax, item in zip(axes, EEM_group_items):

        em_wls, ex_wls, mat = group2mat(item)  # convert group to matrix and extracts ex. wavelengths
        if ex_wls is None:
            raise ValueError(f'Excitation wavelengths of {item.name} could not be extracted from spectra names.')

        x, y = em_wls, ex_wls

        # emission wavelengths limits
        xlim0 = x_lim[0] if x_lim[0] is not None else x[0]
        xlim1 = x_lim[1] if x_lim[1] is not None else x[-1]

        # excitation wavelengths limits
        ylim0 = y_lim[0] if y_lim[0] is not None else y[0]
        ylim1 = y_lim[1] if y_lim[1] is not None else y[-1]

        x_label_down, x_label_top = 'Em. wavelength / nm', 'Em. wavenumber / $10^4$ cm$^{-1}$'
        y_label_left, y_label_right = 'Ex. wavelength / nm', 'Ex. wavenumber / $10^4$ cm$^{-1}$'

        if transform2wavenumber:
            x, y = t2w(x), t2w(y)
            xlim0, xlim1 = t2w(xlim0), t2w(xlim1)
            ylim0, ylim1 = t2w(ylim0), t2w(ylim1)

            # switch the labels
            x_label_down, x_label_top = x_label_top, x_label_down
            y_label_left, y_label_right = y_label_right, y_label_left

        _set_main_axis(ax, xlim=(xlim0, xlim1), ylim=(ylim0, ylim1),
                       y_label=y_label_left,
                       x_label=x_label_down, direction='out',
                       x_major_locator=x_major_locators[0],
                       y_major_locator=y_major_locators[0],
                       x_minor_locator=x_minor_locators[0],
                       y_minor_locator=y_minor_locators[0])

        if log_z:  # use log of z axis
            # mat[mat < 0] = 0
            zmin = mat.max() * 1e-3 if z_lim[0] is None else z_lim[0]  # 3 orders lower than max as default value
        else:
            zmin = mat.min() if z_lim[0] is None else z_lim[0]  # for linear plot, min as default value

        zmax = mat.max() if z_lim[1] is None else z_lim[1]

        # add left axis
        lambda_ax = ax.secondary_xaxis('top', functions=(t2w, t2w))
        lambda_ax.tick_params(which='both', direction='out', zorder=1000)
        if x_major_locators[1] is not None:
            lambda_ax.xaxis.set_major_locator(x_major_locators[1])  # FixedLocator([500, 600, ...])
        if x_minor_locators[1] is not None:
            lambda_ax.xaxis.set_minor_locator(x_minor_locators[1])
        lambda_ax.set_xlabel(x_label_top)

        # add right axis
        lambda_ax2 = ax.secondary_yaxis('right', functions=(t2w, t2w))
        lambda_ax2.tick_params(which='both', direction='out', zorder=1000)
        if y_major_locators[1] is not None:
            lambda_ax2.yaxis.set_major_locator(y_major_locators[1])  # MultipleLocator(20)
        if y_minor_locators[1] is not None:
            lambda_ax2.yaxis.set_minor_locator(y_minor_locators[1])  # AutoMinorLocator(2)
        lambda_ax2.set_ylabel(y_label_right)

        # norm for z values
        norm = mpl.colors.LogNorm(vmin=zmin, vmax=zmax, clip=True) if log_z else mpl.colors.Normalize(vmin=zmin,
                                                                                                       vmax=zmax,
                                                                                                       clip=True)
        _x, _y = np.meshgrid(x, y)
        mappable = ax.pcolormesh(_x, _y, mat.T, norm=norm, cmap=cmap, shading='auto')
        fig.colorbar(mappable, ax=ax, label=z_label, pad=0.17, format=None if log_z else '%.0e')

        if show_title:
            ax.set_title(item.name)

        # if x_major_formatter:
        #     ax_data.xaxis.set_major_formatter(x_major_formatter)
        #     ax_res.xaxis.set_major_formatter(x_major_formatter)

    plt.tight_layout()

    if filepath:
        ext = os.path.splitext(filepath)[1].lower()[1:]
        plt.savefig(fname=filepath, format=ext, transparent=transparent, dpi=dpi)

    plt.show()


def plot_fit(data_item, fit_item, residuals_item, symlog=False, linscale=1, linthresh=100,
              lw_data=0.5, lw_fit=1.5, fig_size_one_graph=(5, 4), y_label='$\\Delta$A', x_label='Time / $\\mu$s',
              x_lim=(None, None), t_mul_factor=1, y_lim=(None, None), x_margin=1, y_margin=1.05, filepath=None, dpi=500,
              transparent=False, x_major_formatter=ScalarFormatter(), x_major_locator=None, y_major_locator=None,
              data_color='red', show_title=True):

    plot_fits([data_item], [fit_item], [residuals_item], n_rows=1, n_cols=1, symlog=symlog, linscale=linscale,
              linthresh=linthresh, lw_data=lw_data, lw_fit=lw_fit, fig_size_one_graph=fig_size_one_graph,
              y_label=y_label, x_label=x_label, x_lim=x_lim, t_mul_factor=t_mul_factor, y_lim=y_lim, x_margin=x_margin,
              y_margin=y_margin, filepath=filepath, dpi=dpi, transparent=transparent,
              x_major_formatter=x_major_formatter, x_major_locator=x_major_locator, y_major_locator=y_major_locator,
              data_color=data_color, show_title=show_title)


def plot_fits(data_group, fit_group, residuals_group, n_rows=None, n_cols=None, symlog=False, linscale=1, linthresh=100,
              lw_data=0.5, lw_fit=1.5, fig_size_one_graph=(5, 4), y_label='$\\Delta$A', x_label='Time / $\\mu$s',
              x_lim=(None, None), t_mul_factor=1, y_lim=(None, None), x_margin=1, y_margin=1.05, filepath=None, dpi=500,
              transparent=False, x_major_formatter=ScalarFormatter(), x_major_locator=None, y_major_locator=None,
              data_color='red', show_title=True):

    n = len(data_group)
    assert n == len(fit_group) == len(residuals_group)

    if n_rows is None and n_cols is None:  # estimate the n_rows and n_cols from the sqrt of number of graphs
        sqrt = n ** 0.5
        n_rows = int(np.ceil(sqrt))
        n_cols = int(sqrt)
    elif n_rows is None and n_cols is not None:
        n_rows = int(np.ceil(n / n_cols))
    elif n_rows is not None and n_cols is None:
        n_cols = int(np.ceil(n / n_rows))

    # assert n_rows * n_cols >= n  # not necessary, if the condition is not valid, fewer plots will be plotted

    fig = plt.figure(figsize=(fig_size_one_graph[0] * n_cols, fig_size_one_graph[1] * n_rows))

    outer_grid = gridspec.GridSpec(n_rows, n_cols, wspace=0.25, hspace=0.3)

    for og, data, fit, res in zip(outer_grid, data_group, fit_group, residuals_group):

        # nice tutorial about gridspec here https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.gridspec.GridSpecFromSubplotSpec.html
        # each unit consist of two graphs - data and residuals
        inner_grid = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=og, wspace=0.1, hspace=0.1,
                                                      height_ratios=(4, 1))

        ax_data = fig.add_subplot(inner_grid[0])
        ax_res = fig.add_subplot(inner_grid[1])

        t_data = data.data[:, 0] * t_mul_factor

        _x_lim = list(x_lim)
        _y_lim = list(y_lim)

        _x_lim[0] = data.data[0, 0] * x_margin * t_mul_factor if _x_lim[0] is None else _x_lim[0]
        _x_lim[1] = data.data[-1, 0] * x_margin * t_mul_factor if _x_lim[1] is None else _x_lim[1]

        _y_lim[0] = data.data[:, 1].min() * y_margin if _y_lim[0] is None else _y_lim[0]
        _y_lim[1] = data.data[:, 1].max() * y_margin if _y_lim[1] is None else _y_lim[1]

        _set_main_axis(ax_data, x_label="", y_label=y_label, xlim=_x_lim, ylim=_y_lim, x_major_locator=x_major_locator,
                       y_major_locator=y_major_locator)
        _set_main_axis(ax_res, x_label=x_label, y_label='res.', xlim=_x_lim, x_minor_locator=None, y_minor_locator=None)

        # plot zero lines
        ax_data.axline((0, 0), slope=0, ls='--', color='black', lw=0.5)
        ax_res.axline((0, 0), slope=0, ls='--', color='black', lw=0.5)

        ax_data.tick_params(labelbottom=False)

        if show_title:
            ax_data.set_title(data.name)
        ax_data.plot(t_data, data.data[:, 1], lw=lw_data, color=data_color)
        ax_data.plot(fit.data[:, 0] * t_mul_factor, fit.data[:, 1], lw=lw_fit, color='black')
        ax_res.plot(res.data[:, 0] * t_mul_factor, res.data[:, 1], lw=lw_data, color=data_color)

        ax_data.set_axisbelow(False)
        ax_res.set_axisbelow(False)

        ax_data.yaxis.set_ticks_position('both')
        ax_data.xaxis.set_ticks_position('both')

        ax_res.yaxis.set_ticks_position('both')
        ax_res.xaxis.set_ticks_position('both')

        if symlog:
            ax_data.set_xscale('symlog', subs=[2, 3, 4, 5, 6, 7, 8, 9], linscale=linscale, linthresh=linthresh)
            ax_res.set_xscale('symlog', subs=[2, 3, 4, 5, 6, 7, 8, 9], linscale=linscale, linthresh=linthresh)
            ax_data.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))
            ax_res.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))

        if x_major_formatter:
            ax_data.xaxis.set_major_formatter(x_major_formatter)
            ax_res.xaxis.set_major_formatter(x_major_formatter)

    if filepath:
        ext = os.path.splitext(filepath)[1].lower()[1:]
        plt.savefig(fname=filepath, format=ext, transparent=transparent, dpi=dpi)

    plt.show()


def reaction_QY_relative(sample_items, actinometer_items, QY_act=1, irradiation_spectrum=None,
                         irradiation_wavelength=None, integration_range=(None, None), V_solution=1,
                         conc_calc_range=(None, None), samples_times=None, actinometers_times=None,
                         c_0_samples=1, c_0_acts=1, integration_method='trapz'):
    """
    TODO.....

    Initial spectrua of both sample and actinometers must be at time = 0 (before irradiaiton).

    if conc_calc_range is (None, None), all spectrum will be taken for calclation of conccentration
    V solution

    integration method works for spectra integration as well as for cumulative time integration


    :param sample_items:
    :param actinometer_items:
    :param irradiation_spectrum:
    :param irradiation_wavelength:
    :param integration_range:
    :param sample_times:
    :param actinometer_times:
    :param integration_method: trapz or simps - simpsons rule
    :return:
    """

    # type checking
    if sample_items is None or actinometer_items is None:
        raise ValueError("Arguments sample_items and actinometer_items must not be None.")

    if not isinstance(sample_items, (list, tuple, SpectrumList)) and \
            not isinstance(actinometer_items, (list, tuple, SpectrumList)):
        raise ValueError("Arguments sample_items and actinometer_items must be type list, tuple or SpectrumList.")

    if isinstance(sample_items, (list, tuple)) and not isinstance(sample_items[0], SpectrumList):
        raise ValueError("Entries of sample_items must be type of SpectrumList.")

    if isinstance(actinometer_items, (list, tuple)) and not isinstance(actinometer_items[0], SpectrumList):
        raise ValueError("Entries of actinometer_items must be type of SpectrumList.")

    if irradiation_wavelength is None and irradiation_spectrum is None:
        raise ValueError("Argument irradiation_spectrum or irradiation_wavelength must be provided.")

    if irradiation_spectrum is not None and irradiation_wavelength is not None:
        raise ValueError("Only one argument of irradiation_spectrum or irradiation_wavelength must be provided.")

    if irradiation_spectrum is None and not isinstance(irradiation_wavelength, (int, float)):
        raise ValueError("Argument irradiation_wavelength must be type of int or float.")

    if irradiation_wavelength is None and not isinstance(irradiation_spectrum, (Spectrum, np.ndarray, list)):
        raise ValueError("Argument irradiation_spectrum must be type of Spectrum, ndarray or list.")

    if not isinstance(integration_range, tuple) or len(integration_range) != 2:
        raise ValueError("Argument integration_range must be type of tuple and have length of 2.")

    samples = [sample_items] if isinstance(sample_items, SpectrumList) else sample_items
    acts = [actinometer_items] if isinstance(actinometer_items, SpectrumList) else actinometer_items

    if irradiation_spectrum:
        irr_sp_x = None
        if isinstance(irradiation_spectrum, Spectrum):
            irr_sp_x = irradiation_spectrum.data[:, 0]
            irr_sp_y = irradiation_spectrum.data[:, 1]
        else:
            irr_sp_y = np.asarray(irradiation_spectrum)

        x0, x1 = integration_range
        if x0 is not None and x1 is not None and x0 > x1:
            x0, x1 = x1, x0

        def abs_photons(data):
            start = 0
            end = data.shape[0]
            start_sp = 0
            end_sp = irr_sp_y.shape[0]
            if x0 is not None:
                start = fi(data[:, 0], x0)
                start_sp = fi(irr_sp_y, x0)
            if x1 is not None:
                end = fi(data[:, 0], x1) + 1
                end_sp = fi(irr_sp_y, x1) + 1

            if start - end != start_sp - end_sp:
                if irr_sp_x is None:
                    raise ValueError("Irradiation spectrum and data does not have equal dimension.")
                irr = np.interp(data[:, 0], irr_sp_x, irr_sp_y)  # interpolate to match data if x vals are provided
            else:
                irr = irr_sp_y[start_sp:end_sp]  # slice the irradiation spectrum to match the data

            x = data[start:end, 0]
            y = data[start:end, 1]

            abs_light = (1 - 10 ** -y) * irr  # (1 - 10 ** -A) * I_irr

            if integration_method == 'trapz':
                return np.trapz(abs_light, x)  # integrate using trapezoidal rule
            else:
                return simps(abs_light, x)  # integrate using simpsons rule

    else:  # only irradiation wavelength
        def abs_photons(data):
            idx = fi(data[:, 0], irradiation_wavelength)
            return 1 - 10 ** -data[idx, 1]  # 1 - 10 ** -A

    _sample_times = [] if samples_times is None else samples_times
    _acs_times = [] if actinometers_times is None else actinometers_times

    if samples_times is None:
        for sample in samples:
            _sample_times = np.asarray([float(sp.name) for sp in sample])

    if actinometers_times is None:
        for act in acts:
            _acs_times = np.asarray([float(sp.name) for sp in act])

    _c_0_samples = [c_0_samples] * len(samples) if isinstance(c_0_samples, (int, float)) else c_0_samples
    _c_0_acts = [c_0_acts] * len(acts) if isinstance(c_0_acts, (int, float)) else c_0_acts

    def cumintegrate(y, x, initial=0):
        if integration_method == 'trapz':
            return cumtrapz(y, x, initial=initial)
        else:
            # simpsons rule
            raise NotImplementedError()  # TODO----->

    def calc_c(unknown_sp_data, sp_data_c0, c0=1):
        """
        Calculation of concentariton by least squares.

        :param unknown_sp_data:
        :param sp_data_c0:
        :param c0:
        :return:
        """
        assert unknown_sp_data.shape == sp_data_c0.shape
        x0, x1 = conc_calc_range
        start = 0
        end = sp_data_c0.shape[0]
        if x0 is not None:
            start = fi(sp_data_c0[:, 0], x0)
        if x1 is not None:
            end = fi(sp_data_c0[:, 0], x1) + 1

        # for min || xA - B ||_2^2 for scalar x, x = sum(A*B) / sum(A*A)
        a = sp_data_c0[start:end, 1]
        b = unknown_sp_data[start:end, 1]

        return c0 * (a * b).sum() / (a * a).sum()

    results_sample = SpectrumList(name='Results of samples')
    results_act = SpectrumList(name='Results of actinometers')

    def calculate_line(sl, times, c0):
        # calculate the amount of absorbed photons
        qs = np.asarray([abs_photons(sp.data) for sp in sl])
        # calculate the time-dependent concentration
        c0s = np.asarray([calc_c(sp.data, sl[0].data, c0) for sp in sl])
        # Delta n = (c(t=0) - c(t)) * V
        d_n_dec = (c0s[0] - c0s) * V_solution
        # cumulative integration of light absorbed: int_0^t q(t') dt'
        np_abs = cumintegrate(qs, times, initial=0)

        return Spectrum.from_xy_values(np_abs, d_n_dec, name=f'Result for {sl.name}')

    for sample, s_times, c0_sample in zip(samples, _sample_times, _c_0_samples):
        results_sample.children.append(calculate_line(sample, s_times, c0_sample))

    for act, act_times, c0_act in zip(acts, _acs_times, _c_0_acts):
        results_act.children.append(calculate_line(act, act_times, c0_act))

    QYs = []

    # calculate for each combination of sample and actinometer kinetics
    for res_sample in results_sample:
        slope_sam, intercept_sam, r_sam, _, err_sam = linregress(res_sample.x, res_sample.y)

        for res_act in results_act:
            slope_act, intercept_act, r_act, _, err_act = linregress(res_act.x, res_act.y)

            # use uncertainty package to automatically propagate errors
            # QY is type of ufloat - uncertainties.core.Variable
            QY = QY_act * ufloat(slope_sam, err_sam) / ufloat(slope_act, err_act)

            QYs.append(QY)

    average_QY = sum(QYs) / len(QYs)

    add_to_list(results_sample)
    add_to_list(results_act)






def bcorr_1D(item, first_der_tresh=1e-4, second_der_tresh=0.1):
    """Gradient based baseline correction"""

    x = item.data[:, 0].copy()
    y = item.data[:, 1].copy()

    grad1 = np.gradient(y, x)  # first central derivative
    grad2 = np.gradient(grad1, x)  # second central derivative

    grad1, grad2 = grad1 / grad1.max(), grad2 / grad2.max()

    zero_idxs = np.argwhere(
        (grad1 < first_der_tresh) & (grad1 > -first_der_tresh) &
        (grad2 < second_der_tresh) & (grad2 >= 0)
    )

    zero_idxs = zero_idxs.squeeze()

    baseline = np.interp(x, x[zero_idxs], y[zero_idxs])

    sp = Spectrum.from_xy_values(x, baseline, f'{item.name} baseline')
    UserNamespace.instance.add_items_to_list(sp)


class UserNamespace:
    instance = None

    def __init__(self, main):

        self.main = main
        UserNamespace.instance = self
        self.tw = self.main.tree_widget

        # execute first commands
        self.main.console.execute_command(
            """
            import numpy as np
            from spectramanipulator.user_namespace import *
            # from spectramanipulator.spectrum import fi, group2mat
            import matplotlib.pyplot as plt
            %matplotlib inline
            
            # setup important methods
            add_to_list = UserNamespace.instance.tw.add_to_list
            load_kinetic = UserNamespace.instance.tw.load_kinetic
            load_kinetics = UserNamespace.instance.tw.load_kinetics
            import_files = UserNamespace.instance.tw.import_files
            
            """
        )

        # from IPython.display import display, Math, Latex\n

        self.main.console.push_variables(
            {
                'main': self.main,
                'tree_widget': self.main.tree_widget,
                'item': self.main.tree_widget.myModel.root
            }
        )

    # def add_items_to_list(self, spectra):
    #     """
    #     Copies all spectra and import them to the treewidget
    #     :param spectra: input parameter can be single spectrum object, or hierarchic list of spectra
    #     """
    #
    #     # self.main.tree_widget.get
    #
    #     if spectra.__class__ == Spectrum:
    #         self.main.tree_widget.import_spectra([spectra])
    #         return
    #
    #     if spectra.__class__.__name__ == 'SpectrumItem':
    #         self.main.tree_widget.import_spectra([spectra.__copy__()])
    #         return
    #
    #     if isinstance(spectra, list):
    #         self.main.tree_widget.import_spectra(spectra)
    #         return
    #
    #     if spectra.__class__.__name__ == 'SpectrumItemGroup' or spectra.__class__.__name__ == 'SpectrumList':
    #         l = []
    #         for sp in spectra:
    #             new_sp = sp.__copy__()
    #             new_sp.group_name = spectra.name
    #             l.append(new_sp)
    #
    #         self.main.tree_widget.import_spectra([l])
    #         return
